#! -*- coding: utf-8 -*-

import glob
import logging
import os
import copy
import os.path

import yaml
import psutil

from brome.core.model.configurator import runner_args_to_dict, get_config_value, default_config
from brome.core.model.meta import *
from brome.core.model.test import Test
from brome.core.model.test_batch import TestBatch
from brome.core.model.test_instance import TestInstance
from brome.core.model.test_result import TestResult
from brome.core.model.utils import *

class BaseRunner(object):
    """Base class for the brome runner

    All brome runner inherit from this base class

    Attributes:
        brome (object)
    """

    def __init__(self, brome):
        self.brome = brome

        self.brome_config_path = self.brome.config_path
        self.commandline_args = self.brome.parsed_args
        self.browsers_config = self.brome.browsers_config

        #CONFIG
        self.config = runner_args_to_dict(self.commandline_args)
        self.brome_config = self.brome.config

        #DATABASE
        setup_database(self.get_config_value('database:*'))

        #Current pid of the runner
        current_pid = os.getpid()

        #Get the activated tests list
        self.tests = self.get_activated_tests()

        #Create test batch
        session = Session()

        self.starting_timestamp = datetime.now()
        sa_test_batch = TestBatch(starting_timestamp = self.starting_timestamp)
        sa_test_batch.pid = current_pid
        sa_test_batch.total_tests = len(self.tests)

        session.add(sa_test_batch)

        session.commit()

        self.test_batch_id = sa_test_batch.id

        session.close()

        #RUNNER LOG DIR
        self.root_test_result_dir = self.get_config_value("project:test_batch_result_path")

        if self.root_test_result_dir:
            self.runner_dir = os.path.join(
                self.root_test_result_dir,
                "tb_%s"%self.test_batch_id
            )
            self.relative_runner_dir = "tb_%s"%self.test_batch_id
            create_dir_if_doesnt_exist(self.runner_dir)
        else:
            self.runner_dir = False

        #LOGGING
        self.configure_logger()

        #SCREENSHOT CACHE
        if self.get_config_value('runner:cache_screenshot'):
            #Dictionary that contains all the screenshot name
            self.screenshot_cache = {}

    def kill_pid(self, pid):
        """Kill process by pid

        Args:
            pid (int)
        """
        try:

            p = psutil.Process(pid)

            p.terminate()

            self.info_log('Killed [pid:%s][name:%s]'%(p.pid, p.name()))
        except psutil.NoSuchProcess:
            self.error_log('No such process: [pid:%s]'%pid)

    def kill(self, procname):
        """Kill by process name

        Args:
            procname (str)
        """
        for proc in psutil.process_iter():
            if proc.name() == procname:
                self.info_log('[pid:%s][name:%s] killed'%(proc.pid, proc.name()))
                proc.kill()

    def get_available_tests(self, search_query = None, test_name = None):
        """Return a list of all the available tests

        This function might call exit(1) if not test were found

        Returns:
            list
        """

        available_tests = []

        script_folder_name = self.get_config_value('project:script_folder_name')

        if search_query:

            tests_path = os.path.join(
                self.get_config_value('project:absolute_path'), 
                script_folder_name,
                'test_%s.py'%search_query
            )
            tests = sorted(glob.glob(tests_path))

            for test in tests:
                module_test = test.split(os.sep)[-1][:-3]
                available_tests.append(__import__('%s.%s'%(script_folder_name, module_test), fromlist = ['']))

        elif test_name:
            if test_name.endswith('.py'):
                test_name = test_name[:-3]

            try:
                available_tests.append(__import__('%s.%s'%(script_folder_name, test_name), fromlist = ['']))
            except ImportError:
                pass

        if not len(available_tests):
            query = search_query if search_query else test_name
            print "No script found with the provided query: %s"%query
            exit(1)
        
        return available_tests

    def get_activated_tests(self):
        """Return a list of all the activated tests

        The behavior of this function is determined by the received CLI arguments

        Returns:
            list
        """
        tests = []

        #test file
        if self.get_config_value('runner:test_file'):
            test_file_path = self.get_config_value('runner:test_file')
            with open(test_file_path, 'r') as f:
                test_list = yaml.load(f)

            for test in test_list:
                tests.append(self.get_available_tests(test)[0])

        elif self.get_config_value('runner:test_name'):
            tests = self.get_available_tests(test_name = self.get_config_value('runner:test_name'))
        else:
            test_search_query = self.get_config_value('runner:test_search_query')

            #by index or slice e.g.: [0:12], [:], [0], [-1]
            if test_search_query.find('[') != -1:
                exec('tests = self.get_available_tests("*")%s'%test_search_query)

            #by name
            else:
                tests = self.get_available_tests(test_search_query)

        return tests

    def get_config_value(self, config_name):
        """Return the effective config value

        Args:
            config_name (str)

        Returns:
            value (the type vary according to the config_name)
        """

        config_list = [
            self.config,
            self.brome_config
        ]
        value = get_config_value(config_list,config_name)

        return value

    def configure_logger(self):
        """Configure the test batch runner logger
        """

        logger_name = 'brome_runner'

        self.logger = logging.getLogger(logger_name)

        format_ = self.get_config_value("logger_runner:format")

        #Stream logger 
        if self.get_config_value('logger_runner:streamlogger'):
            sh = logging.StreamHandler()
            stream_formatter = logging.Formatter(format_)
            sh.setFormatter(stream_formatter)
            self.logger.addHandler(sh)

        #File logger
        if self.get_config_value('logger_runner:filelogger'):
            fh = logging.FileHandler(os.path.join(self.runner_dir, '%s.log'%logger_name))
            file_formatter = logging.Formatter(format_)
            fh.setFormatter(file_formatter)
            self.logger.addHandler(fh)

        self.logger.setLevel(getattr(logging, self.get_config_value('logger_runner:level')))

    def print_test_summary(self, executed_tests):
        """Print test summary

        When the test batch is finished a test summary will be printed

        Args:
            executed_tests (list)
        """

        separator = '---------------------'

        session = Session()

        sa_test_batch = session.query(TestBatch).filter(TestBatch.id == self.test_batch_id).one()

        #TITLE
        self.info_log('******* TEST BATCH SUMMARY ********')

        #TOTAL NUMBER OF EXECUTED TESTS
        base_query = session.query(TestResult).filter(TestResult.test_batch_id == self.test_batch_id)
        total_test = base_query.count()
        total_test_successful = base_query.filter(TestResult.result == True).count()
        total_test_failed = base_query.filter(TestResult.result == False).count()
        self.info_log('Total_test: %s; Total_test_successful: %s; Total_test_failed: %s'%(total_test, total_test_successful, total_test_failed))

        #EXECUTION TIME
        self.info_log("Total execution time: %s"%(sa_test_batch.ending_timestamp - sa_test_batch.starting_timestamp))

        #SEPARATOR
        self.info_log(separator)

        self.info_log('Failed tests')

        #FAILED TESTS
        failed_test_list = []
        for test_result in sa_test_batch.test_results:
            if not test_result.result and not test_result.test in failed_test_list:
                query = session.query(Test).filter(Test.id == test_result.test_id)
                if query.count():
                    test = query.one()
                    failed_test_list.append(test.test_id)
                    self.info_log("[%s] %s"%(test.test_id, test.name))
                else:
                    failed_test_list.append(test_result.title)
                    self.info_log("[noid] %s"%(test_result.title))

        if not failed_test_list:
            self.info_log('No test failed!')
        
        #SEPARATOR
        self.info_log(separator)

        #TEST INSTANCE REPORT
        for test in executed_tests:
            #TITLE
            self.info_log('%s %s'%(test._name, test.pdriver.get_id()))

            sa_test_instance = session.query(TestInstance).filter(TestInstance.id == test._test_instance_id).one()

            #TEST EXECUTION TIME
            try:
                self.info_log("Test execution time: %s"%(sa_test_instance.ending_timestamp - sa_test_instance.starting_timestamp))
            except TypeError:
                self.info_log("Test execution time exception")

            #TEST INSTANCE SUMMARY
            results = test.get_test_result_summary()
            for result in results:
                self.info_log(result)

            #CRASH REPORT
            if test._crash_error:
                self.info_log(test._crash_error)
            else:
                self.info_log('No crash!')

            #SEPARATOR
            self.info_log(separator)

        #END
        self.info_log('Finished')

    def get_logger_dict(self):
        return {'batchid': self.test_batch_id}

    def debug_log(self, msg):
        self.logger.debug("[debug]%s"%msg, extra=self.get_logger_dict())

    def info_log(self, msg):
        self.logger.info("%s"%msg, extra=self.get_logger_dict())

    def warning_log(self, msg):
        self.logger.warning("[warning]%s"%msg, extra=self.get_logger_dict())

    def error_log(self, msg):
        self.logger.error("[error]%s"%msg, extra=self.get_logger_dict())

    def critical_log(self, msg):
        self.logger.critical("[critical]%s"%msg, extra=self.get_logger_dict())



#! -*- coding: utf-8 -*-

from flask_login import UserMixin

from brome.core.model.utils import *
from brome.webserver.extensions import bcrypt
from brome.core.model.meta import SurrogatePK, Base, Column, DateTime, Text, Boolean

class User(SurrogatePK, UserMixin, Base):

    username = Column(Text())
    email = Column(Text())
    password = Column(Text())
    created_at = Column(DateTime())

    def __init__(self, username, email, password):
        self.username = username
        self.email = email
        self.created_at = datetime.now()
        self.set_password(password)

    def set_password(self, password):
        self.password = bcrypt.generate_password_hash(password)

    def check_password(self, value):
        return bcrypt.check_password_hash(self.password, value)

    @property
    def full_name(self):
        return "{0} {1}".format(self.first_name, self.last_name)

    def __repr__(self):
        return '<User({username!r})>'.format(username=self.username)



#! -*- coding: utf-8 -*-

from brome.core.model.meta import SurrogatePK, Base, Column, DateTime, Boolean, Text, Integer, ForeignKey

class TestResult(SurrogatePK, Base):
    result = Column(Boolean())
    timestamp = Column(DateTime())
    browser_id = Column(Text())
    screenshot_path = Column(Text())
    videocapture_path = Column(Text())
    extra_data = Column(Text())
    title = Column(Text())

    test_id = Column(Integer, ForeignKey('test.id'))
    test_instance_id = Column(Integer, ForeignKey('testinstance.id'))
    test_batch_id = Column(Integer, ForeignKey('testbatch.id'))



# -*- coding: utf-8 -*-
"""Public section, including homepage and signup."""
from flask import (Blueprint, request, render_template, flash, url_for,
                    redirect, session)
from flask.ext.login import login_user, login_required, logout_user

from brome.core.model.user import User
from brome.webserver.extensions import login_manager
from brome.webserver.public.forms import LoginForm
from brome.webserver.admin.forms import RegisterForm
from brome.webserver.utils import flash_errors
from brome.webserver.extensions import db

blueprint = Blueprint('public', __name__, static_folder="../static")

@login_manager.user_loader
def load_user(id):
    return db.session.query(User).get(int(id))

@blueprint.route("/", methods=["GET", "POST"])
def home():
    form = LoginForm(request.form)
    # Handle logging in
    if request.method == 'POST':
        if form.validate_on_submit():
            login_user(form.user)
            flash("You are logged in.", 'success')
            redirect_url = request.args.get("next") or url_for("public.home")
            return redirect(redirect_url)
        else:
            flash_errors(form)
    return render_template("public/home.html", form=form)

@blueprint.route('/logout/')
@login_required
def logout():
    logout_user()
    flash('You are logged out.', 'info')
    return redirect(url_for('public.home'))

@blueprint.route("/register/", methods=['GET', 'POST'])
def register():
    
    closed_registration = blueprint.app.config.get('CLOSED_REGISTRATION', False)

    form = RegisterForm(request.form, csrf_enabled=False, app = blueprint.app)
    if form.validate_on_submit():
        new_user = User(username=form.username.data,
                        email=form.email.data,
                        password=form.password.data
                        )

        db.session.add(new_user)
        db.session.commit()

        flash("Thank you for registering. You can now log in.", 'success')

        return redirect(url_for('public.home'))
    else:
        flash_errors(form)

    return render_template('public/register.html', form=form, closed_registration = closed_registration)

@blueprint.route("/about/")
def about():
    form = LoginForm(request.form)
    return render_template("public/about.html", form=form)



#! -*- coding: utf-8 -*-

from brome.core.model.meta import SurrogatePK, Text, Base, Column, DateTime, ForeignKey, Integer, relationship

class TestInstance(SurrogatePK, Base):

    name = Column(Text())

    starting_timestamp = Column(DateTime())
    ending_timestamp = Column(DateTime())
    extra_data = Column(Text())

    test_batch_id = Column(Integer, ForeignKey('testbatch.id'))
    test_results = relationship("TestResult", backref="testinstance")



#! -*- coding: utf-8 -*-

import os
from glob import glob
import shutil
import psutil
from datetime import datetime
import json

from IPython import embed
from brome.core.model.test_batch import TestBatch
from brome.core.model.test_instance import TestInstance
from brome.core.model.test_result import TestResult
from brome.core.model.test_crash import TestCrash
from brome.core.model.test import Test
from brome.webserver.extensions import db

def analyse_network_capture(app, network_capture_path, analyse_function):
    network_capture_path = os.path.join(app.brome.get_config_value('project:test_batch_result_path'), network_capture_path)

    analyse_network_capture_func = app.brome.get_config_value("webserver:%s"%analyse_function)
    module_name = analyse_network_capture_func.split(':')[0]
    function_name = analyse_network_capture_func.split(':')[1]

    module = __import__(module_name, fromlist = [''])
    data = getattr(module, function_name)(network_capture_path)
    
    return data

def get_network_capture(app, testbatch_id):
    data = []

    relative_dir = os.path.join(
        "tb_%s"%testbatch_id,
        "network_data"
    )

    abs_dir = os.path.join(
        app.brome.get_config_value('project:test_batch_result_path'),
        relative_dir
    )

    if os.path.isdir(abs_dir):
        for f in os.listdir(abs_dir):
            network_capture = {}
            network_capture['name'] = f.split('.')[0]
            network_capture['path'] = os.path.join(relative_dir, f)
            analyse = app.brome.get_config_value("webserver:analyse_network_capture_func")
            if analyse:
                network_capture['analyse'] = analyse

            data.append(network_capture)

    return data

def get_test_batch_list():
    data = db.session.query(TestBatch).order_by(TestBatch.id.desc()).all()

    return data

def get_test_instance_list(testbatch_id):
    test_instances = db.session.query(TestInstance)\
                            .filter(TestInstance.test_batch_id == testbatch_id)\
                            .all()

    for test_instance in test_instances:
        extra_data = json.loads(test_instance.extra_data)
        test_instance.private_ip = extra_data.get('instance_private_ip', 'Unknown')
        test_instance.private_dns = extra_data.get('instance_private_dns', 'Unknown')
        test_instance.public_ip = extra_data.get('instance_public_ip', 'Unknown')
        test_instance.public_dns = extra_data.get('instance_public_dns', 'Unknown')

        if test_instance.ending_timestamp is None:
            test_instance.running_info = "Is running..."
        else:
            test_instance.running_info = "Total execution time: %s"%(test_instance.ending_timestamp - test_instance.starting_timestamp)

    return test_instances

def get_total_execution_time(app, testbatch_id):
    test_batch = get_test_batch(testbatch_id)

    if test_batch.ending_timestamp:
        total_execution_time = test_batch.ending_timestamp - test_batch.starting_timestamp
    else:
        total_execution_time = 'still running...'

    return total_execution_time

def get_test_batch_result_dir(app, testbatch_id):
    return os.path.join(
        app.brome.get_config_value("project:test_batch_result_path"),
        'tb_%s'%testbatch_id
    )

def delete_test_batch(app, testbatch_id):
    test_batch = get_test_batch(testbatch_id)

    db.session.delete(test_batch)
    db.session.commit()

    shutil.rmtree(get_test_batch_result_dir(app, testbatch_id))

def stop_test_batch(app, testbatch_id):
    test_batch = get_test_batch(testbatch_id)

    test_batch.killed = True
    ret = psutil.pid_exists(test_batch.pid)
    if not ret and test_batch.ending_timestamp is None:
        test_batch.ending_timestamp = datetime.now()

    db.session.commit()

def get_test_batch(testbatch_id):
    data = db.session.query(TestBatch).filter(TestBatch.id == testbatch_id).one()

    return data

def get_test_instance(test_instance_id):
    data = db.session.query(TestInstance).filter(TestInstance.id == test_instance_id).one()

    return data

def get_test_batch_detail(app, testbatch_id):
    data = db.session.query(TestBatch).filter(TestBatch.id == testbatch_id).one()

    data.total_crashes = get_total_crashes(app, testbatch_id)
    data.total_executing_tests = get_total_executing_tests(testbatch_id)
    data.total_finished_tests = get_total_finished_tests(testbatch_id)
    data.total_tests = get_total_tests(testbatch_id)
    data.total_screenshots = get_test_batch_screenshot(app, testbatch_id, only_total = True)
    data.total_test_results = get_test_batch_test_result(app, testbatch_id, only_total = True)
    data.total_failed_tests = get_test_batch_test_result(app, testbatch_id, only_failed_total = True)
    data.total_execution_time = get_total_execution_time(app, testbatch_id)

    return data

def get_total_tests(testbatch_id):
    return get_test_batch(testbatch_id).total_tests

def get_total_crashes(app, testbatch_id):
    relative_logs_dir = os.path.join(
        "tb_%s"%testbatch_id,
        "crashes"
    )

    abs_logs_dir = os.path.join(
        app.brome.get_config_value('project:test_batch_result_path'),
        relative_logs_dir
    )

    return len(glob(os.path.join(abs_logs_dir, '*.log')))

def get_total_finished_tests(testbatch_id):
    count = db.session.query(TestInstance)\
        .filter(TestInstance.test_batch_id == testbatch_id)\
        .filter(TestInstance.ending_timestamp != None)\
        .count()

    return count

def get_total_executing_tests(testbatch_id):
    count = db.session.query(TestInstance)\
        .filter(TestInstance.test_batch_id == testbatch_id)\
        .filter(TestInstance.ending_timestamp == None)\
        .count()

    return count

def get_test_list(app):
    data = []

    tests_dir = os.path.join(
        app.brome.get_config_value('project:absolute_path'),
        "tests"
    )

    if os.path.isdir(tests_dir):
        tests = glob(os.path.join(tests_dir, 'test_*.py'))
        for test in sorted(tests):
            name = test.split(os.sep)[-1][len('test_'):-3]
            data.append({'name': name})

    return data

def get_browser_list(app):
    data = []

    for key, browser in app.brome.browsers_config.iteritems():
        if browser.get('available_in_webserver', False):
            browser_config = {}
            browser_config['id'] = key
            browser_config['name'] = browser['browserName']

            icon = "fa-%s"%browser_config['name'].lower().replace(' ', '-')
            if browser_config.has_key('deviceName'):
                if 'android' in browser_config['deviceName'].lower():
                    icon = "fa-android"
                elif 'iphone' in browser_config['deviceName'].lower():
                    icon = "fa-mobile"
                elif 'ipad' in browser_config['deviceName'].lower():
                    icon = "fa-tablet"

            browser_config['icon'] = icon

            data.append(browser_config)

    return data

def get_test_batch_screenshot(app, testbatch_id, only_total = False):
    data = []

    relative_dir = os.path.join(
        "tb_%s"%testbatch_id,
        "screenshots"
    )

    abs_dir = os.path.join(
        app.brome.get_config_value('project:test_batch_result_path'),
        relative_dir
    )

    if os.path.isdir(abs_dir):
        for browser_dir in os.listdir(abs_dir):
            screenshot_list = os.listdir(os.path.join(abs_dir, browser_dir))

            for screenshot in screenshot_list:
                data.append({
                    'title': screenshot.split('.')[0].replace('_', ' '),
                    'browser_id': browser_dir.replace('_', ' '),
                    'path': os.path.join(relative_dir, browser_dir, screenshot)
                })

    if only_total:
        return len(data)

    return data

def get_test_result(app, test_result_id):
    return db.session.query(TestResult)\
                .filter(TestResult.id == test_result_id)\
                .one()

def get_test_crash(app, test_crash_id):
    return db.session.query(TestCrash)\
                .filter(TestCrash.id == test_crash_id)\
                .one()

def get_test_batch_test_result(app, testbatch_id, only_total = False, only_failed_total = False):
    query_ = db.session.query(TestResult)\
                .filter(TestResult.test_batch_id == testbatch_id)

    if only_total:
        return query_.count()
    elif only_failed_total:
        return query_.filter(TestResult.result == False).count()
    else:
        query_ = query_.outerjoin(Test, TestResult.test_id == Test.id)
        return query_.order_by(TestResult.result, Test.test_id).all()

def get_test_batch_log(app, testbatch_id):
    abs_logs_dir = os.path.join(
        app.brome.get_config_value('project:test_batch_result_path'),
        "tb_%s"%testbatch_id
    )

    runner_log = []
    log_path = os.path.join(abs_logs_dir, "brome_runner.log")
    if os.path.isfile(log_path):
        with open(log_path, 'r') as f:
            runner_log = f.read().replace('"', '&quot;').replace("'", '&quot;').splitlines()

    return runner_log

def get_test_batch_test_instance_log(app, testbatch_id, index):
    data = []

    relative_logs_dir = os.path.join(
        "tb_%s"%testbatch_id,
        "logs"
    )

    abs_logs_dir = os.path.join(
        app.brome.get_config_value('project:test_batch_result_path'),
        relative_logs_dir
    )

    #def ls_(dir_):
    #    """
    #    ctime = lambda f: os.stat(os.path.join(dir_, f)).st_ctime
    #    return [f for f in sorted(os.listdir(dir_), key=ctime)]
    #    """
    #    return [f for f in os.listdir(dir_)]

    if os.path.isdir(abs_logs_dir):
        for log in sorted(os.listdir(abs_logs_dir)):
            data.append({
                'name': log,
                'path': os.path.join(relative_logs_dir, log)
            })

    return data[index:]

def get_test_batch_crashes(app, testbatch_id):
    result = db.session.query(TestCrash)\
                        .filter(TestCrash.test_batch_id == testbatch_id)\
                        .all()

    for crash in result:
        crash.trace_list = crash.trace.split(os.linesep)

    return result

def get_test_batch_video_recording(app, testbatch_id, only_total = False):
    data = []

    relative_dir = os.path.join(
        "tb_%s"%testbatch_id,
        "video_recording"
    )

    abs_dir = os.path.join(
        app.brome.get_config_value('project:test_batch_result_path'),
        relative_dir
    )

    if os.path.isdir(abs_dir):
        for browser_dir in sorted(os.listdir(abs_dir)):
            video_recording_list = os.listdir(os.path.join(abs_dir, browser_dir))

            for video_recording in video_recording_list:
                data.append({
                    'title': '%s - %s'%(video_recording.split('.')[0].replace('_', ' '), browser_dir.replace('_', ' ')),
                    'path': os.path.join(relative_dir, browser_dir, video_recording)
                })

    if only_total:
        return len(data)

    return data



#! -*- coding: utf-8 -*-

import traceback

from selenium.webdriver.common.touch_actions import TouchActions

from brome.core.model.utils import *

class ProxyElement(object):
    
    def __init__(self, element, selector, pdriver):
        self._element = element
        self.selector = selector
        self.pdriver = pdriver
    
    def __getattr__(self, funcname):
        return getattr(self._element, funcname)

    def error_log(self, msg):
        self.pdriver.error_log(u"[%s] %s"%(self.__repr__(), msg))

    def debug_log(self, msg):
        self.pdriver.debug_log(u"[%s] %s"%(self.__repr__(), msg))

    def __repr__(self):
        msg = [u"WebElement (selector: '%s')"%self.selector]

        try:
            if self._element.get_attribute('id'):
                msg.append(u"(id: '%s')"%self._element.get_attribute('id'))
            
            if self._element.get_attribute('name'):
                msg.append(u"(name: '%s')"%self._element.get_attribute('name'))

            if self._element.get_attribute('class'):
                msg.append(u"(class: '%s')"%self._element.get_attribute('class'))

        except Exception as e:
            self.pdriver.debug_log("exception in __repr__: %s"%unicode(e))

        return u' '.join(msg)

    def is_displayed(self, **kwargs):
        self.debug_log(u"Is displayed")

        raise_exception = kwargs.get(
                                    'raise_exception',
                                    self.pdriver.get_config_value(
                                        'proxy_driver:raise_exception'
                                    )
                                )
        retry = kwargs.get('retry', True)

        try:
            is_displayed = self._element.is_displayed()
            self.debug_log(u"is displayed")
            return is_displayed
        except StaleElementReferenceException:
            if retry:
                #NOTE this is an imperfect solution since we can have found the element with find_last
                #TODO find a better way to handle this edge case
                self.debug_log(u"StaleElementReferenceException; retrying...")
                self._element = self.pdriver.find(self.selector, raise_exception = False)
                if self._element:
                    return self.is_displayed(retry = False)
                else:
                    return False
            else:
                if raise_exception:
                    raise
                else:
                    return False

    def click(self, **kwargs):
        self.debug_log(u"Clicking on element")
        
        highlight = kwargs.get( 
                            'highlight',
                            self.pdriver.get_config_value(
                                'highlight:highlight_when_element_is_clicked'
                            )
                    )
        wait_until_clickable = kwargs.get(
                            'wait_until_clickable',
                            self.pdriver.get_config_value(
                                'proxy_element:wait_until_clickable'
                            )
                    )

        if wait_until_clickable:
            #TODO manage the raise exception better
            self.pdriver.wait_until_clickable(self.selector, raise_exception = True)

        if highlight:
            self.highlight(
                style = self.pdriver.get_config_value(
                            'highlight:style_when_element_is_clicked'
                        )
                )

        def _click():
            if self.pdriver.get_config_value("proxy_element:use_touch_instead_of_click"):
                touch_action = TouchActions(self.pdriver._driver)
                touch_action.tap(self._element).perform()
            else:
                self._element.click()

        try:
            _click()
        except (InvalidElementStateException, WebDriverException) as e:
            self.debug_log(u"click exception: %s"%unicode(e))
            sleep(2)
            self.scroll_into_view()
            _click()
        except StaleElementReferenceException as e:
            self.debug_log(u"click exception StaleElementReferenceException: %s"%unicode(e))
            sleep(2)
            self._element = self.pdriver.find(self.selector)
            _click()

        return True

    def send_keys(self, value, **kwargs):
        self.debug_log(u"Sending keys")

        highlight = kwargs.get( 
                            'highlight',
                            self.pdriver.get_config_value(
                                'highlight:highlight_when_element_receive_keys'
                            )
                    )
        wait_until_clickable = kwargs.get(
                            'wait_until_clickable',
                            self.pdriver.get_config_value(
                                'proxy_element:wait_until_clickable'
                            )
                    )

        if wait_until_clickable:
            #TODO manage the raise exception better
            self.pdriver.wait_until_clickable(self.selector, raise_exception = True)

        if highlight:
            self.highlight(
                style = self.pdriver.get_config_value(
                            'highlight:style_when_element_receive_keys'
                        )
            )

        clear = kwargs.get('clear', False)

        if clear:
            self.clear()

        try:
            self._element.send_keys(value)
        except StaleElementReferenceException as e:
            self.debug_log(u"send_keys exception StaleElementReferenceException: %s"%unicode(e))
            sleep(2)
            self._element = self.pdriver.find(self.selector)
            self._element.send_keys(value)
        except (InvalidElementStateException, WebDriverException) as e:
            self.debug_log(u"send_keys exception: %s"%unicode(e))
            sleep(2)
            self._element.send_keys(value)

        return True

    def clear(self):
        self.debug_log(u"Clearing element")

        try:
            self._element.clear()
        except (InvalidElementStateException, WebDriverException) as e:
            self.debug_log(u"send_keys exception: %s"%unicode(e))
            sleep(2)
            self._element.clear()
        except StaleElementReferenceException as e:
            self.debug_log(u"send_keys exception StaleElementReferenceException: %s"%unicode(e))
            sleep(2)
            self._element = self.pdriver.find(self.selector)
            self._element.clear()

        return True

    def highlight(self, **kwargs):
        self.debug_log(u"Highlighting element")
        """
            kwargs:
                style: css
                highlight_time: int; default: .3
        """
        style = kwargs.get('style')
        highlight_time = kwargs.get('highlight_time', .3)

        driver = self._element._parent

        try:
            original_style = self._element.get_attribute('style')

            driver.execute_script("arguments[0].setAttribute('style', arguments[1]);", self._element, style)
        except StaleElementReferenceException:
            return False

        sleep(highlight_time)

        try:
            driver.execute_script("arguments[0].setAttribute('style', arguments[1]);", self._element, original_style)
        except StaleElementReferenceException:
            return False

        return True

    def scroll_into_view(self, **kwargs):
        self.debug_log(u"Scrolling into view element")

        raise_exception = kwargs.get(
                                    'raise_exception',
                                    self.pdriver.get_config_value(
                                        'proxy_driver:raise_exception'
                                    )
                                )
        try:
            self.pdriver.execute_script("arguments[0].scrollIntoView()", self._element)

        except WebDriverException as e:
            if raise_exception:
                raise
            else:
                tb = traceback.format_exc()
                self.error_log(u'scroll_into_view WebDriverException: %s'%unicode(tb))
                return False

        return True

    def select_all(self, **kwargs):
        self.debug_log(u"Selecting all in element")

        raise_exception = kwargs.get(
                                    'raise_exception',
                                    self.pdriver.get_config_value(
                                        'proxy_driver:raise_exception'
                                    )
                                )

        #http://stackoverflow.com/questions/985272/selecting-text-in-an-element-akin-to-highlighting-with-your-mouse
        try:
            self.pdriver.execute_script("""
                var element = arguments[0],
                    range, selection;

                if (document.body.createTextRange) {
                    range = document.body.createTextRange();
                    range.moveToElementText(element);
                    range.select();
                } else if (window.getSelection) {
                    selection = window.getSelection();        
                    range = document.createRange();
                    range.selectNodeContents(element);
                    selection.removeAllRanges();
                    selection.addRange(range);
                }
            """, self._element)
        except WebDriverException:
            if raise_exception:
                raise
            else:
                tb = traceback.format_exc()
                self.error_log(u'select_all WebDriverException: %s'%unicode(tb))
                return False

        return True




import os

from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy import engine_from_config, MetaData, create_engine
from decorator import decorator

from brome.core.model import *
from brome.core.model.meta import MultipleResultsFound

engine = None

Session = scoped_session(sessionmaker())

class Base(object):

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

Base = declarative_base(cls=Base)

# establish a constraint naming convention.
# see http://docs.sqlalchemy.org/en/latest/core/constraints.html#configuring-constraint-naming-conventions
#
Base.metadata.naming_convention={
        "pk": "pk_%(table_name)s",
        "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
        "uq": "uq_%(table_name)s_%(column_0_name)s",
        "ix": "ix_%(table_name)s_%(column_0_name)s"
    }

def delete_database(sqlalchemy_url):
    if sqlalchemy_url.startswith('sqlite'):
        db_path = sqlalchemy_url.split('/')[-1]
        try:
            os.remove(db_path)
            print 'Database (%s) deleted!'%db_path
        except OSError:
            pass
    else:
        db_name = sqlalchemy_url.split('/')[-1]
        engine = create_engine('/'.join(sqlalchemy_url.split('/')[:-1]))
        conn = engine.connect()
        conn.execute("DROP DATABASE IF EXISTS %s"%db_name)
        conn.close()
        print 'Database (%s) deleted!'%db_name

def create_database(sqlalchemy_url):
    if sqlalchemy_url.startswith('sqlite'):
        engine = create_engine(sqlalchemy_url)
        print 'Database (%s) created!'%sqlalchemy_url.split('/')[-1]
    else:
        db_name = sqlalchemy_url.split('/')[-1]
        engine = create_engine('/'.join(sqlalchemy_url.split('/')[:-1]))
        conn = engine.connect()
        conn.execute("CREATE DATABASE %s"%db_name)
        conn.close()
        print 'Database (%s) created!'%db_name

def setup_database(config):
    """Setup the application given a config dictionary."""

    global engine
    global Base
    engine = engine_from_config(config, "sqlalchemy.")
    Session.configure(bind=engine)
    Base.metadata.create_all(engine)

def update_test(session, test_dict):
    from brome.core.model.test import Test
    print 'Updating the test'
    for test_id, test_config in test_dict.iteritems():
        if type(test_config) == dict:
            name = test_config['name']
        else:
            name = test_config

        #NEW TEST
        if not session.query(Test).filter(Test.test_id == test_id).count():
            test = Test(test_id = test_id, name = name)
            session.add(test)
            print 'Added test id', test_id
        #UPDATE TEST
        else:
            test = session.query(Test).filter(Test.test_id == test_id).one()
            if test.name != name:
                test.name = name
                print 'Updated test id', test_id

    session.commit()

    print 'Done!'



#! -*- coding: utf-8 -*-

from urlparse import urlparse
import string
import os
import traceback
from datetime import datetime
import sys
from subprocess import call, Popen
from time import sleep

from selenium.common.exceptions import *
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from IPython import embed
from pudb import set_trace
import psutil

from .exceptions import *

def get_timestamp():
    return datetime.now().strftime('%Y-%m-%d-%H-%M-%S')

def string_to_filename(s):
    valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
    filename = ''.join(c for c in s if c in valid_chars)
    filename = filename.replace(' ','_')
    return filename

def say(msg):
    if sys.platform in ['win32', 'linux2']:
        call(["espeak", msg])
    else:
        call(["say", msg])

def create_dir_if_doesnt_exist(path):
    if not os.path.exists(path):
        os.makedirs(path)

if sys.platform == 'win32':
    devnull = open('log-null', 'w')
else:
    devnull = open('/dev/null', 'w')

def kill_by_pid(pid):
    p = psutil.Process(pid)
    p.terminate()
    print '[pid:%s]killed'%pid

def kill_by_name(procname):
    for proc in psutil.process_iter():
        if proc.name() == procname:
            print '[pid:%s][name:%s] killed'%(proc.pid, proc.name())
            proc.kill()

def kill_by_found_string_in_cmdline(procname, string):
    for proc in psutil.process_iter():
        if proc.name() == procname:
            for cmd in proc.cmdline():
                if cmd.find(string) != -1:
                    print '[pid:%s][name:%s] killed'%(proc.pid, proc.name())
                    proc.kill()



# -*- coding: utf-8 -*-

from flask_assets import Bundle, Environment

css = Bundle(
    "libs/jQueryui/jquery-ui.css",
    "libs/primeui/theme/bootstrap/theme.css",
    "libs/primeui/primeui-2.0-min.css",
    "libs/primeui/css/datagrid/datagrid.css", #http://forum.primefaces.org/viewtopic.php?f=16&t=42486
    "libs/bootstrap/dist/css/bootstrap.css",
    "css/style.css",
    "libs/bootstrap_dialog/bootstrap-dialog.min.css",
    "libs/videojs/video-js.min.css",
    filters="cssmin",
    output="public/css/all.css"
)

js = Bundle(
    "libs/jQuery/dist/jquery.js",
    "libs/jQueryui/jquery-ui.js",
    "libs/primeui/primeui-2.0-hacked.js", #hacked version of primeui
    "libs/primeui/js/datagrid/datagrid.js", #http://forum.primefaces.org/viewtopic.php?f=16&t=42486
    "libs/bootstrap/dist/js/bootstrap.js",
    "libs/videojs/video.min.js",
    "js/plugins.js",
    "libs/bootstrap_dialog/bootstrap-dialog.min.js",
    filters='jsmin',
    output="public/js/all.js"
)

assets = Environment()

assets.register("js_all", js)
assets.register("css_all", css)



import socket
import threading
import traceback
import os
from subprocess import Popen
from datetime import datetime
import math

import virtualbox

from brome.core.model.utils import *
from brome.core.model.meta.base import Session
from brome.core.runner.base_runner import BaseRunner
from brome.core.runner.ec2_instance import EC2Instance
from brome.core.runner.virtualbox_instance import VirtualboxInstance
from brome.core.runner.localhost_instance import LocalhostInstance
from brome.core.runner.saucelabs_instance import SauceLabsInstance
from brome.core.runner.browser_config import BrowserConfig
from brome.core.model.test_batch import TestBatch

class GridRunner(BaseRunner):
    """GridRunner run on Ec2, Virtualbox, Browserstack, saucelabs and also localhost
    """
    def __init__(self, *args):
        super(GridRunner, self).__init__(*args)

        self.selenium_pid = None
        self.instances_ip = {}
        self.instances = {}
        self.alive_instances = []
        self.xvfb_pids = []
        self.browser_configs = {}

    def execute(self):
        """Execute the test batch
        """

        #START SELENIUM GRID
        if self.get_config_value('grid_runner:start_selenium_server'):
            self.start_selenium_server()

        #Get all the browsers id
        self.browsers_id = self.get_config_value('runner:remote_runner').split(',')

        #Start all the instances
        instance_threads = []
        for i, browser_id in enumerate(self.browsers_id):
            browser_config = BrowserConfig(
                runner = self,
                browser_id = browser_id,
                browsers_config = self.browsers_config
            )

            self.browser_configs[browser_id] = browser_config

            #EC2
            if browser_config.location == 'ec2':

                max_number_of_instance = browser_config.get('max_number_of_instance')
                nb_browser_by_instance = browser_config.get('nb_browser_by_instance')

                if len(self.tests) < \
                    max_number_of_instance * \
                    nb_browser_by_instance:

                    nb_instance_to_launch = int(math.ceil(float(len(self.tests))/nb_browser_by_instance))

                else:
                    nb_instance_to_launch = max_number_of_instance

                for j in range(nb_instance_to_launch):
                    ec2_instance = EC2Instance(
                        runner = self,
                        browser_config = browser_config,
                        index = j
                    )

                    ec2_instance_thread = InstanceThread(ec2_instance)
                    ec2_instance_thread.start()

                    instance_threads.append(ec2_instance_thread)

            #VIRTUALBOX
            elif browser_config.location == 'virtualbox':
                #Instanciate only one vbox
                if not hasattr(self, 'vbox'):
                    self.vbox = virtualbox.VirtualBox()

                vbox_instance = VirtualboxInstance(
                    runner = self,
                    browser_config = browser_config,
                    index = i,
                    vbox = self.vbox
                )

                vbox_instance_thread = InstanceThread(vbox_instance)
                vbox_instance_thread.start()

                instance_threads.append(vbox_instance_thread)

            #SAUCELABS
            elif browser_config.location == 'saucelabs':

                if not self.instances.get(browser_id):
                    self.instances[browser_id] = []

                self.instances[browser_id].append(SauceLabsInstance())

            #BROWSERSTACK
            elif browser_config.location == 'browserstack':

                if not self.instances.get(browser_id):
                    self.instances[browser_id] = []

                self.instances[browser_id].append(BrowserstackInstance())

            #LOCALHOST
            elif browser_config.location == 'localhost':

                max_number_of_instance = browser_config.get('max_number_of_instance', 1)

                if len(self.tests) < max_number_of_instance:
                    nb_instance_to_launch = len(self.tests)
                else:
                    nb_instance_to_launch = max_number_of_instance

                for i in range(0, nb_instance_to_launch):
                    if not self.instances.get(browser_id):
                        self.instances[browser_id] = []

                    self.instances[browser_id].append(LocalhostInstance(self, browser_config, test_name = i))

        for t in instance_threads:
            t.join()

        self.info_log("The test batch is now ready!")

        try:
            self.run()
        except:
            tb = traceback.format_exc()
            self.error_log("Exception in run of the grid runner: %s"%str(tb))
            raise

        finally:
            try:
                self.tear_down_instances()

                #Kill selenium server
                if self.get_config_value('grid_runner:kill_selenium_server'):
                    if self.selenium_pid:
                        self.kill_pid(self.selenium_pid)

                #Kill xvfb process
                for xvfb_pid in self.xvfb_pids:
                    self.kill_pid(xvfb_pid)

            except:
                tb = traceback.format_exc()
                self.error_log("Exception in finally block of the grid runner: %s"%str(tb))
                    
        self.info_log("The test batch is finished.")

    def resolve_instance_by_ip(self, ip):
        """Resolve an instance by his ip
        """
        return self.instances_ip[ip]

    def run(self):
        """Run all the test in the test batch
        """

        executed_tests = []
        try:

            active_thread = 0
            start_thread = True
            current_index = 0
            active_thread_by_browser_id = {}

            test_index_by_browser_id = {}
            for browser_id in self.browsers_id:
                test_index_by_browser_id[browser_id] = 0

            self.kill_test_batch_if_necessary()
            
            while active_thread or start_thread:
                start_thread = False

                self.kill_test_batch_if_necessary()

                for browser_id in self.browsers_id:

                    if active_thread_by_browser_id.get(browser_id) is None:
                        active_thread_by_browser_id[browser_id] = 0
                    else:
                        current_active_thread = 0
                        for thread in threading.enumerate():

                            #Make sure that the thread is TestThread
                            if hasattr(thread, 'test'):
                                if thread.test._browser_config.browser_id == browser_id:
                                    current_active_thread += 1

                        active_thread_by_browser_id[browser_id] = current_active_thread

                    for j in range(0, len(self.instances[browser_id]) - active_thread_by_browser_id[browser_id]):

                        self.kill_test_batch_if_necessary()

                        if test_index_by_browser_id[browser_id] < len(self.tests):
                            current_index += 1

                            test = self.tests[test_index_by_browser_id[browser_id]]

                            test_index_by_browser_id[browser_id] += 1

                            test_ = test.Test(
                                runner = self,
                                test_batch_id = self.test_batch_id,
                                browser_config = self.browser_configs[browser_id],
                                name = test.Test.name,
                                index = test_index_by_browser_id[browser_id]
                            )
                            test_.pdriver.embed_disabled = True

                            thread = TestThread(test_)
                            thread.start()

                            executed_tests.append(test_)

                active_thread = threading.active_count() - 1
                if active_thread:
                    try:
                        active_thread_test_number = len([tn for tn in threading.enumerate() if type(tn) != threading._MainThread and hasattr(tn, 'test')])
                        self.info_log("Active thread number: %s"%active_thread_test_number)
                        self.info_log("Active thread name: %s"%(', '.join([
                                "%s-%s"%(
                                    th.test._browser_config.browser_id,
                                    th.test._name
                                ) for th in threading.enumerate() if type(th) != threading._MainThread and hasattr(th, 'test')
                            ])
                        ))
                    except Exception as e:
                        self.error_log("print active exception: %s"% str(e))

                #TIMEOUT
                if (self.starting_timestamp - datetime.now()).total_seconds() >\
                    self.get_config_value("grid_runner:max_running_time"):

                    self.error_log("max_running_time reached... terminating!")
                    raise TestRunnerKilledException()

                self.kill_test_batch_if_necessary()

                sleep(10)

        except TestRunnerKilledException:
            pass

        except Exception:
            tb = traceback.format_exc()
            self.error_log("Run exception: %s"%str(tb))

        session = Session()
        sa_test_batch = Session.query(TestBatch).filter(TestBatch.id == self.test_batch_id).one()
        sa_test_batch.ending_timestamp = datetime.now()
        session.commit()
        session.close()

        self.print_test_summary(executed_tests)

    def kill_test_batch_if_necessary(self):
        """Kill the test batch

        If the test_batch.killed is set to true then the test batch will be kill
        """
        session = Session()
        sa_test_batch = Session.query(TestBatch).filter(TestBatch.id == self.test_batch_id).one()
        session.close()

        if sa_test_batch.killed:
            self.info_log("Killing itself")
            for t in [t for t in threading.enumerate() if type(t) != threading._MainThread]:
                self.info_log("Killing: %s"%t.test._name)
                t.test.end()

            raise TestRunnerKilledException("Killed")

    def tear_down_instances(self):
        """Tear down all instances
        """

        self.info_log('Tearing down all instances...')

        for instance in self.alive_instances:
            instance.tear_down()

        self.info_log('[Done]Tearing down all instances')

    def start_selenium_server(self):
        """Start the selenium server
        """

        ip = self.get_config_value("grid_runner:selenium_server_ip")
        port = self.get_config_value("grid_runner:selenium_server_port")

        def is_selenium_server_is_running():
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            result = s.connect_ex((ip, port))
            s.close()
            return not result

        if not is_selenium_server_is_running():
            self.info_log('Starting selenium server...')

            command = self.get_config_value("grid_runner:selenium_server_command")\
                .format(
                    **self.get_config_value("grid_runner:*")
                )

            self.info_log('Selenium hub command: %s'%command)

            process = Popen(
                command.split(' '),
                stdout=open(os.path.join(self.runner_dir, "hub.log"), 'a'),
                stderr=open(os.path.join(self.runner_dir, "hub.log"), 'a'),
            )

            self.selenium_pid = process.pid

            self.info_log('Selenium server pid: %s'%self.selenium_pid)
        else:
            self.info_log('Selenium is already running.')
            return True

        for i in range(30):
            self.info_log('Waiting for the selenium server to start...')
            result = is_selenium_server_is_running()

            if result:
                self.info_log('[Done]Selenium server is running.')
                return True
            sleep(2)

        raise Exception("Selenium server did not start!")

class InstanceThread(threading.Thread):
    """Theard that start the instance
    """

    def __init__(self, instance):
        threading.Thread.__init__(self)
        self.instance = instance
        self.runner = self.instance.runner

    def run(self):
        success = False
        try:
            success = self.instance.startup()
        except Exception as e:
            self.runner.critical_log("Exception in InstanceThread instance startup: %s"%unicode(e))

        if not self.runner.instances.get(self.instance.browser_config.browser_id):
            self.runner.instances[self.instance.browser_config.browser_id] = []

        if success:
            self.runner.instances[self.instance.browser_config.browser_id].append(self.instance)
            self.runner.instances_ip[self.instance.get_ip()] = self.instance

        self.runner.alive_instances.append(self.instance)

class TestThread(threading.Thread):
    """Thread that run the test
    """

    def __init__(self, test):
        threading.Thread.__init__(self)
        self.test = test

    def run(self):
        self.test.execute()

class TestRunnerKilledException(Exception):
    """Exception that is raised when the test batch is killed
    """
    pass



#! -*- coding: utf-8 -*-

import virtualbox
import paramiko
import netifaces as ni

from brome.core.model.utils import *
from .base_instance import BaseInstance

class VirtualboxInstance(BaseInstance):
    """Virtual box instance
    
    Attributes:
        runner (object)
        browser_config (object)
        index (int)

    Kwargs:
        vbox (object)
    """

    def __init__(self, runner, browser_config, index, **kwargs):
        self.runner = runner
        self.browser_config = browser_config
        self.index = index

        self.vbox = kwargs.get('vbox')
        
    def get_ip(self):
        """Return the ip address of the node
        """

        return self.ip

    def execute_command(self, command, **kwargs):
        """Execute a command on the node

        Args:
            command (str)

        Kwargs:
            username (str)
        """

        self.info_log("executing command: %s"%command)

        try:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            username = kwargs.get('username', self.browser_config.get('username'))
            password = self.browser_config.get('password')

            ssh.connect(self.ip, username = username, password = password)

            stdin, stdout, stderr = ssh.exec_command(command)

            output = stdout.read()

            ssh.close()

            return (stdout, stderr)

        except Exception as e:
            msg = "Execute_command exception: %s"%str(e)
            self.error_log(msg)
            raise Exception(msg)

    def startup(self):
        """This will launch and configure the virtual box machine
        """

        #Do not launch the virtual machine
        if not self.browser_config.get('launch', False):
            return

        self.info_log("Starting up")

        try:
            #Cleanup the vbox guestproperty variable
            call(['VBoxManage', 'guestproperty', 'delete', self.browser_config.get('vbname'), 'wait_until_ready'])
            call(['VBoxManage', 'guestproperty', 'delete', self.browser_config.get('vbname'), 'hub_ip'])

            #Find the machine
            vm = self.vbox.find_machine(self.browser_config.get('vbname'))

            #Session
            self.session = virtualbox.Session()

            #Determine if the vm is running or not
            #Launch it is necessary
            vm_already_running = False
            if str(vm.session_state) != 'Locked':
                vm.launch_vm_process(self.session, self.browser_config.get('vbox_type', 'gui'), '')
            else:
                vm_already_running = True

            user_session = vm.create_session()
            #NOTE this is another way to record the session
            #now castro is used
            #user_session.machine.video_capture_enabled = False 

            #User session
            self.vbox_session = user_session

            #Wait until the vm is ready
            if not vm_already_running:

                instance_ready = False
                #TODO should be configurable
                timeout = 60

                self.info_log('Waiting for instance to start...')

                for i in range(timeout):
                    try:
                        value, timestamp, flags = user_session.machine.get_guest_property('wait_until_ready')

                        if value == u'ready':
                            instance_ready = True
                            break

                    except virtualbox.library.VBoxErrorInvalidVmState:
                        pass

                    sleep(1)

                sleep(3)
                if instance_ready:
                    self.info_log('[Done] Instance ready...')
                else:
                    raise Exception("Timeout error: the virtualbox machine is still not ready.")

            #HUB IP
            hub_ip = ni.ifaddresses('en0')[2][0]['addr']

            self.info_log("Hub ip: %s"%hub_ip)

            #VM IP
            vm_ip = vm.get_guest_property("/VirtualBox/GuestInfo/Net/0/V4/IP")[0]
            self.ip = vm_ip

            self.info_log("vm ip: %s"%vm_ip)

            #Start selenium on the node
            if not vm_already_running:

                #LINUX
                if self.browser_config.get('platform').lower() == "linux":

                    self.info_log('Starting the selenium node server')

                    #Update the hub_ip browser config
                    self.browser_config.config['hub_ip'] = hub_ip

                    command = self.browser_config.get("selenium_command").format(**self.browser_config.config)
                    self.execute_command(command)

                #WINDOWS
                elif self.browser_config.get('platform').lower() == "windows":

                    self.info_log("Setting the guest property in Windows")

                    user_session.machine.set_guest_property("hub_ip", "%s:%s"%(hub_ip, '4444'), '')

            return True

        except Exception as e:
            self.error_log('Exception: %s'%e)
            raise
        
    def tear_down(self):
        """Tear down the virtual box machine
        """

        if not self.browser_config.get('terminate'):
            self.warning_log("Skipping terminate")
            return

        self.info_log("Tearing down")

        if self.browser_config.get('platform').lower() == 'linux':
            self.execute_command("shutdown -h now", username = 'root')

        elif self.browser_config.get('platform').lower() == 'windows':
            self.session.console.power_down()

    def get_id(self):
        return '%s - %s'%(self.browser_config.browser_id, self.index)

    def debug_log(self, msg):
        self.runner.debug_log("[%s]%s"%(self.get_id(), msg))

    def info_log(self, msg):
        self.runner.info_log("[%s]%s"%(self.get_id(), msg))

    def warning_log(self, msg):
        self.runner.warning_log("[%s]%s"%(self.get_id(), msg))

    def error_log(self, msg):
        self.runner.error_log("[%s]%s"%(self.get_id(), msg))

    def critial_log(self, msg):
        self.runner.critial_log("[%s]%s"%(self.get_id(), msg))



# -*- coding: utf-8 -*-

import tempfile
import os
import logging

from flask import Flask, render_template
import flask_sijax

from brome.core.model import *
from brome.core.model.meta.base import Base
from brome.webserver.assets import assets
from brome.webserver.extensions import (
    bcrypt,
    cache,
    db,
    login_manager,
    debug_toolbar,
)
from brome.webserver import public, admin, testbatch
from brome.core.model.utils import create_dir_if_doesnt_exist

log = logging.getLogger('werkzeug')
log.setLevel(logging.ERROR)

def create_app(brome):
    app = Flask(__name__)
    app.brome = brome

    app.config.update(brome.get_config_value("webserver:*"))

    app.config['SQLALCHEMY_DATABASE_URI'] = brome.get_config_value('database:sqlalchemy.url')
    app.config["SIJAX_STATIC_PATH"] = os.path.join('.', os.path.dirname(__file__), 'static/libs/sijax/')
    app.config["SIJAX_JSON_URI"] = '/static/libs/sijax/json2.js'

    app.temp_path = tempfile.gettempdir()

    register_extensions(app)
    register_blueprints(app)
    register_errorhandlers(app)

    @app.before_first_request
    def setup():
        Base.metadata.create_all(bind=db.engine)

    return app

def configure_logging(app):
    logger_name = 'brome_webserver'

    app.logger = logging.getLogger(logger_name)

    #Stream logger 
    if app.brome.get_config_value('webserver:streamlogger'):
        sh = logging.StreamHandler()
        app.logger.addHandler(sh)

    #File logger
    if app.brome.get_config_value('webserver:filelogger'):
        fh = logging.FileHandler(os.path.join(app.runner_dir, '%s.log'%logger_name))
        app.logger.addHandler(fh)

    brome.logger.setLevel(getattr(logging, self.get_config_value('webserver:logger_level')))

def register_extensions(app):
    assets.init_app(app)

    bcrypt.init_app(app)

    cache.init_app(app)

    db.init_app(app)

    login_manager.init_app(app)

    debug_toolbar.init_app(app)

    flask_sijax.Sijax(app)

def register_blueprints(app):
    app.register_blueprint(public.views.blueprint)
    public.views.blueprint.app = app

    app.register_blueprint(admin.views.blueprint)
    admin.views.blueprint.app = app

    app.register_blueprint(testbatch.views.blueprint)
    testbatch.views.blueprint.app = app

def register_errorhandlers(app):

    def render_error(error):
        # If a HTTPException, pull the `code` attribute; default to 500
        error_code = getattr(error, 'code', 500)

        return render_template("{0}.html".format(error_code)), error_code

    for errcode in [401, 404, 500]:
        app.errorhandler(errcode)(render_error)



# -*- coding: utf-8 -*-

from flask_wtf import Form
from wtforms import TextField, PasswordField
from wtforms.validators import DataRequired, Email, EqualTo, Length
from IPython import embed

from brome.webserver.extensions import db
from brome.core.model.configurator import default_config, load_brome_config, save_brome_config
from brome.core.model.user import User

class RegisterForm(Form):
    username = TextField('Username',
                    validators=[DataRequired(), Length(min=3, max=25)])
    email = TextField('Email',
                    validators=[DataRequired(), Email(), Length(min=6, max=40)])
    password = PasswordField('Password',
                                validators=[DataRequired(), Length(min=6, max=40)])
    confirm = PasswordField('Verify password',
                [DataRequired(), EqualTo('password', message='Passwords must match')])
    token = PasswordField('Token')

    def __init__(self, *args, **kwargs):
        super(RegisterForm, self).__init__(*args, **kwargs)
        self.app = kwargs.get('app')
        self.user = None

    def validate(self):
        initial_validation = super(RegisterForm, self).validate()
        if not initial_validation:
            return False

        if self.app.config.get('CLOSED_REGISTRATION', False):
            if self.token.data != self.app.config.get('REGISTRATION_TOKEN'):
                self.token.errors.append("The registration security token doesn't match")
                return False

        user = db.session.query(User).filter_by(username = self.username.data).first()
        if user:
            self.username.errors.append("Username already registered")
            return False

        user = db.session.query(User).filter_by(email = self.email.data).first()
        if user:
            self.email.errors.append("Email already registered")
            return False

        return True

class ConfigForm(object):
    def __init__(self, app):
        self.app = app

        self.fields = self.get_fields()

    def get_fields(self):
        data = default_config
        config = load_brome_config(self.app.brome.config_path)
        
        for section_key, section_item in config.iteritems():

            for option_key, option_item in section_item.iteritems():

                #User defined section
                if not data.has_key(section_key):
                    data[section_key] = {}
                    data[section_key][option_key] = {
                        'value': option_item,
                        'type': 'input',
                        'visible': True,
                        'title': option_key
                    }

                else:
                    #User defined option
                    if not data[section_key].has_key(option_key):
                        data[section_key][option_key] = {
                            'value': option_item,
                            'type': 'input',
                            'visible': True,
                            'title': option_key
                        }
                    else:
                        data[section_key][option_key]['value'] = option_item

        return data

    def save(self, data):
        for section_id, section in self.fields.iteritems():
            for field_id, field in section.iteritems():
                try:
                    value = data.getlist("%s_%s"%(section_id, field_id))[0]
                    if value == 'on':
                        field['value'] = True
                    else:
                        field['value'] = value
                except IndexError:
                    field['value'] = False

        config = {}
        for k,v in self.fields.iteritems():
            for ik, iv in v.iteritems():
                if not config.has_key(k):
                    config[k] = {}

                if not config[k].has_key(ik):
                    config[k][ik] = {}

                if iv['type'] == 'number':
                    config[k][ik] = int(iv['value'])
                else:
                    config[k][ik] = iv['value']

        return save_brome_config(self.app.brome.config_path, config) 



#! -*- coding: utf-8 -*-

class BaseInstance(object):

    def get_ip(self):
        pass

    def execute_command(self, command):
        pass

    def startup(self):
        pass

    def tear_down(self):
        pass

    def debug_log(self, msg):
        print msg

    def info_log(self, msg):
        print msg

    def warning_log(self, msg):
        print msg

    def error_log(self, msg):
        print msg

    def critial_log(self, msg):
        print msg



# -*- coding: utf-8 -*-

"""Extensions module. Each extension is initialized in the app factory located
in app.py
"""

from flask_bcrypt import Bcrypt
bcrypt = Bcrypt()

from flask_login import LoginManager
login_manager = LoginManager()

from flask_sqlalchemy import SQLAlchemy
db = SQLAlchemy()

from flask_cache import Cache
cache = Cache()

from flask_debugtoolbar import DebugToolbarExtension
debug_toolbar = DebugToolbarExtension()



# -*- coding: utf-8 -*-
"""Helper utilities and decorators."""
from flask import flash


def flash_errors(form, category="warning"):
    """Flash all errors for a form."""
    for field, errors in form.errors.items():
        for error in errors:
            flash("{0} - {1}"
                  .format(getattr(form, field).label.text, error), category)



#! -*- coding: utf-8 -*-

import re
from StringIO import StringIO

from lxml import etree
from lxml.cssselect import CSSSelector
from cssselect.parser import SelectorSyntaxError

class Selector(object):
    """Selector class to manage the selector brome system

    Args:
        pdriver (object)
        selector (str)
    """
    def __init__(self, pdriver, selector):
        self._pdriver = pdriver
        self._selector = selector

        #List of selector support
        if not type(selector) is list:
            self._selector_list= [self._selector]
        else:
            self._selector_list = self._selector

        self._effective_selector_list = self.resolve_selector()
        self._effective_selector_type = self.resolve_selector_type()

        self.find_function, self.find_by = self.resolve_function()

        self._effective_selector = ''.join([self._get_selector(sel) for sel in self._effective_selector_list])

        #XPATH VALIDATION
        if self._effective_selector_type == 'xpath':
            if self._pdriver.get_config_value("proxy_driver:validate_xpath_selector"):
                xpath_test = etree.parse(StringIO('<foo><bar></bar></foo>'))
                try:
                    xpath_test.xpath(self.get_selector())
                except etree.XPathEvalError, e:
                    raise Exception("Invalid xpath: %s"%self.get_selector())

        #CSS VALIDATION
        elif self._effective_selector_type == 'css':
            if self._pdriver.get_config_value("proxy_driver:validate_css_selector"):
                try:
                    CSSSelector(self.get_selector())
                except SelectorSyntaxError:
                    raise Exception("Invalid css selector: %s"%self.get_selector())

    def __repr__(self):
        return "Selector: effective selector (%s), raw selector (%s)"%(self.get_selector(), self._selector)

    #GET
    def get_selector(self):
        """Get the final selector

        This selector can be feed to selenium
        """
        return self._effective_selector

    def _get_selector(self, selector):
        """Get the selector without his brome specific prefix
        """
        return selector[3:]

    def get_type(self, selector):
        """Get the type of the selector

        see SELECTOR_DICT for supported type
        """
        try:
            return SELECTOR_DICT[selector[:3]]
        except KeyError:
            raise Exception("""
                All selector need to start with either:
                    'nm:' (name), 'xp:' (xpath), 'cn:' (classname), 'id:' (id), 'cs:' (css), 'tn:' (tag name), 'lt:' (link text), 'pl:' (partial link text)
            """)

    #RESOLVE
    def resolve_selector(self):
        """Resolve the selector variable in place
        """
        effective_selector_list = []

        for current_selector in self._selector_list:

            #INLINE SELECTOR
            if self.get_type(current_selector) != 'selector_variable':
                effective_selector_list.append(current_selector)

            #SELECTOR VARIABLE
            else:
                #Make sure the proxy driver have a selector dictionary
                if self.get_type(current_selector) == 'selector_variable':
                    if not hasattr(self._pdriver, 'selector_dict'):
                        raise Exception("You must provide a selector dictionary if you want to use the selector variable type")

                #Make sure that the selector dictionary contains the selector variable
                if not self._pdriver.selector_dict.has_key(self._get_selector(current_selector)):
                    raise Exception("Cannot find the selector variable (%s) in the selector dictionary"%self._get_selector(current_selector))

                effective_selector = self._pdriver.selector_dict[self._get_selector(current_selector)]
                if type(effective_selector) is dict:
                    current_browser_id = False

                    keys = [key for key in effective_selector.keys() if key != 'default']
                    for key in keys:
                        for target in key.split('|'):
                            try:
                                re.search(target, self._pdriver.get_id()).group(0)
                                current_browser_id = key
                            except AttributeError:
                                pass

                    if current_browser_id:
                        effective_selector_list.append(effective_selector.get(current_browser_id))
                    else:
                        effective_selector_list.append(effective_selector.get('default'))

                else:
                    if self.get_type(effective_selector) in [value for key,value in SELECTOR_DICT.items() if key != 'selector_variable']:
                        effective_selector_list.append(effective_selector)
                    else:
                        raise Exception("""
                            All selector need to start with either:
                                'nm:' (name), 'xp:' (xpath), 'cn:' (classname), 'id:' (id), 'cs:' (css), 'tn:' (tag name), 'lt:' (link text), 'pl:' (partial link text)
                        """)

        return effective_selector_list

    def resolve_selector_type(self):
        """Resolve the selector type

        This make sure that all the selectors provided are of the same type (in case of a list of selector)
        """
        resolved_selector_type_list = []
        for current_selector in self._effective_selector_list:
            resolved_selector_type_list.append(self.get_type(current_selector))

        set_ = set(resolved_selector_type_list)

        #VALIDATE THAT ALL SELECTOR ARE OF THE SAME TYPE
        if len(set_) != 1:
            raise Exception("If you provide a list of selector then all selectors must be of the same type")
        else:
            return set_.pop()

    def resolve_function(self):
        """Resolve the selenium function that will be use to find the element
        """
        selector_type = self._effective_selector_type

        #NAME
        if selector_type == 'name':
            return ('find_elements_by_name', 'NAME')

        #XPATH
        elif selector_type == 'xpath':
            return ('find_elements_by_xpath', 'XPATH')

        #CLASSNAME
        elif selector_type == 'class_name':
            return ('find_elements_by_class_name', 'CLASS_NAME')

        #ID
        elif selector_type == 'id':
            return ('find_element_by_id', 'ID')

        #CSS
        elif selector_type == 'css':
            return ('find_elements_by_css_selector', 'CSS_SELECTOR')

        #TAGNAME
        elif selector_type == 'tag_name':
            return ('find_elements_by_tag_name', 'TAG_NAME')

        #LINK TEXT
        elif selector_type == 'link_text':
            return ('find_elements_by_link_text', 'LINK_TEXT')

        #PARTIAL LINK TEXT
        elif selector_type == 'partial_link_text':
            return ('find_elements_by_partial_link_text', 'PARTIAL_LINK_TEXT')

SELECTOR_DICT = {
    'nm:': 'name',
    'xp:': 'xpath',
    'lt:': 'link_text',
    'pl:': 'partial_link_text',
    'cn:': 'class_name',
    'id:': 'id',
    'cs:': 'css',
    'tn:': 'tag_name',
    'sv:': 'selector_variable'
}



#! -*- coding: utf-8 -*-

import json
import logging
import os.path
import os
import pickle
from urlparse import urlparse

from selenium.webdriver.common.proxy import *
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from castroredux import CastroRedux

from brome.core.model.utils import *
from brome.core.model.stateful import Stateful
from brome.core.model.proxy_driver import ProxyDriver
from brome.core.model.meta.base import Session
from brome.core.model.test_instance import TestInstance
from brome.core.model.test_result import TestResult
from brome.core.model.test_crash import TestCrash
from brome.core.model.test import Test
from brome.core.model.configurator import get_config_value, parse_brome_config_from_browser_config, default_config, test_config_to_dict

class BaseTest(object):
    """Base test class used to init the driver, load and save state, start and stop video recording

    Attributes:
        runner (object): the runner that created the BaseTest instance
        name (str): the name of the test
        index (int): the index of the test
        browser_config (object): the browser config; used to init the driver
        test_batch_id (int): the test batch id
    """

    def __init__(self, runner, name, index, browser_config, test_batch_id, **kwargs):
        self._runner = runner
        self._name = name
        self._index = index
        self._browser_config = browser_config
        self._test_batch_id = test_batch_id
        self._localhost_instance = kwargs.get('localhost_instance')

        self._crash_error = False

        #TEST BATCH DIRECTORY
        if self._runner.runner_dir:
            self._runner_dir = self._runner.runner_dir
        else:
            self._runner_dir = False

        #LOGGING
        self.configure_logger()

        #DRIVER
        self.pdriver = self.init_driver()

        #ASSIGN THE TEST NAME TO THE INSTANCE
        if self._browser_config.location in ['ec2', 'virtualbox']:
            self._runner.resolve_instance_by_ip(self.pdriver.get_ip_of_node()).testname = self._name

        #TEST RESULT DIRECTORY
        self.configure_test_result_dir()

        #DRIVER RESOLUTION
        self.pdriver.configure_resolution()

        #VIDEO RECORDING
        self.start_video_recording()

        #TEST KWARGS
        self._test_config = test_config_to_dict(self.get_config_value("runner:test_config"))

        session = Session()

        extra_data = {} 
        if self._browser_config.location == 'ec2':
            private_ip = self.pdriver.get_ip_of_node()
            extra_data['instance_private_ip'] = self._runner.instances_ip[private_ip].private_ip
            extra_data['instance_public_ip'] = self._runner.instances_ip[private_ip].public_ip
            extra_data['instance_public_dns'] = self._runner.instances_ip[private_ip].public_dns
            extra_data['instance_private_dns'] = self._runner.instances_ip[private_ip].private_dns

        sa_test_instance = TestInstance(
            starting_timestamp = datetime.now(),
            name = self._name,
            test_batch_id = self._test_batch_id,
            extra_data = json.dumps(extra_data)
        )

        session.add(sa_test_instance)
        session.commit()

        self._test_instance_id = sa_test_instance.id

        session.close()

    def start_video_recording(self):
        self._video_capture_file_relative_path = False

        if self._browser_config.get('record_session'):
            node_ip = self.pdriver.get_ip_of_node()

            self._video_capture_file_path = os.path.join(
                self._video_recording_dir,
                string_to_filename('%s.flv'%(self._name.replace(' ', '_')))
            )

            self._video_capture_file_relative_path = os.path.join(
                self._video_recording_relative_dir,
                string_to_filename('%s.flv'%(self._name.replace(' ', '_')))
            )

            self._video_recorder = CastroRedux(
                self._video_capture_file_path,
                host = node_ip,
                port = self._browser_config.get('vnc_port', 5900)
            )

            try:
                self._video_recorder.start()
                self.info_log("CastroRedux started (ip: %s)(output: %s)"%(node_ip, self._video_capture_file_path))
            except Exception as e:
                self.info_log("CastroRedux exception: %s"%str(e))

    def stop_video_recording(self):
        if hasattr(self, '_video_recorder'):
            self.info_log("Finalizing the video capture...")

            #Let the time to the driver to quit so we have the full picture
            sleep(5)

            try:
                self._video_recorder.stop()
            except Exception as e:
                tb = traceback.format_exc()
                self.error_log("CastroRedux error traceback: %s"%str(e))
            """
            file_name = "%s/%s"%(self.video_capture_dir, self.config.get('name').replace(' ', '_'))
            Popen(["/usr/bin/ffmpeg", "-i", "%s.flv"%file_name, "-vcodec", "libvpx", "-acodec", "libvorbis", "%s.webm"%file_name], stdout=devnull, stderr=devnull)
            """

    def init_driver(self, retry = 10):
        """Init driver will instanciate a webdriver according to the browser config

        First a webdriver is instanciate according to the provided browser config
        then the webdriver is wrapped in the brome proxy_driver

        Args:
            retry (int): the number of time the init driver will retry to init the driver if it fail
                            default: 10

        Returns:
            proxy_driver instance
            
        Raises:
            Exception InitDriverException
            Exception InvalidBrowserName

        """
        #LOCAL
        if self._browser_config.location == 'localhost':
            if self._browser_config.get('browserName').lower() in ['chrome', 'firefox', 'phantomjs'] \
                and self._browser_config.get('enable_proxy'):
                mitm_proxy = "localhost:%s"%self._localhost_instance.proxy_port

                proxy = Proxy({
                    'proxyType': ProxyType.MANUAL,
                    'httpProxy': mitm_proxy,
                    'sslProxy': mitm_proxy
                })

                #NOTE http://www.seleniumhq.org/docs/04_webdriver_advanced.jsp#using-a-proxy
                #FIREFOX
                if self._browser_config.get('browserName').lower() == 'firefox':
                    profile  = webdriver.FirefoxProfile()
                    profile.set_proxy(proxy = proxy)
                    driver = webdriver.Firefox(firefox_profile=profile)

                #CHROME
                elif self._browser_config.get('browserName').lower() == 'chrome':
                    chrome_options = webdriver.ChromeOptions()
                    chrome_options.add_argument("--proxy-server={0}".format(proxy))
                    driver = webdriver.Chrome(chrome_options = chrome_options)

                #PHANTOMJS
                elif self._browser_config.get('browserName').lower() == 'phantomjs':

                    #TODO investigate why this doesnt work in phantomjs
                    """
                    #WAY 1
                    #http://stackoverflow.com/questions/14699718/how-do-i-set-a-proxy-for-phantomjs-ghostdriver-in-python-webdriver
                    service_args = [
                    '--proxy=127.0.0.1:%s'%self._localhost_instance.proxy_port,
                    '--proxy-type=http',
                    ]
                    driver = webdriver.PhantomJS(service_args=service_args)

                    #WAY 2
                    desired_cap = self._browser_config.config
                    proxy.add_to_capabilities(desired_cap)
                    driver = webdriver.PhantomJS(desired_capabilities = desired_cap)
                    """
                    raise NotImplemented()

                #NOTE: AFAIK Safari doesn't support proxy
                #https://code.google.com/p/selenium/issues/detail?id=5051
            else:
                try:
                    driver = getattr(webdriver, self._browser_config.get('browserName'))()
                except AttributeError:
                    raise InvalidBrowserName("The browserName('%s') is invalid"%self._browser_config.get('browserName'))

        #APPIUM
        elif self._browser_config.location == 'appium':
            driver = webdriver.Remote(
                command_executor='http://127.0.0.1:4723/wd/hub',
                desired_capabilities = self._browser_config.config
            )

        #SAUCELABS
        elif self._browser_config.location == 'saucelabs':
            driver = webdriver.Remote(
                command_executor='http://%s:%s@ondemand.saucelabs.com:80/wd/hub'%(
                    self.get_config_value("saucelabs:username"),
                    self.get_config_value("saucelabs:key")
                ),
                desired_capabilities = self._browser_config.config
            )

        #BROWSERSTACK
        elif self._browser_config.location == 'browserstack':
            driver = webdriver.Remote(
                command_executor='http://%s:%s@hub.browserstack.com:80/wd/hub'%(
                    self.get_config_value("browserstack:username"),
                    self.get_config_value("browserstack:key")
                ),
                desired_capabilities = self._browser_config.config
            )

        #REMOTE
        elif self._browser_config.location in ['virtualbox', 'ec2']:
            config = self._browser_config.config

            desired_cap = {}
            desired_cap['browserName'] = config.get('browserName')
            desired_cap['platform'] = config.get('platform')
            desired_cap['javascriptEnabled'] = True

            if config.get('enable_proxy'):
                mitm_proxy = "localhost:%s"%config.get('proxy_port', 8080)

                proxy = Proxy({
                    'proxyType': ProxyType.MANUAL,
                    'httpProxy': mitm_proxy,
                    'sslProxy': mitm_proxy
                })

            if desired_cap['browserName'].lower() == "chrome":
                chrome_options = Options()
                chrome_options.add_argument("--test-type")
                chrome_options.add_argument("--disable-application-cache")

                if config.get('enable_proxy'):
                    chrome_options.add_argument("--proxy-server={0}".format(proxy))

                desired_cap=chrome_options.to_capabilities()

            try:
                command_executor = "http://%s:%s/wd/hub"%(
                    self.get_config_value("grid_runner:selenium_server_ip"),
                    self.get_config_value("grid_runner:selenium_server_port")
                )

                if desired_cap['browserName'].lower() == "firefox" and config.get('enable_proxy'):
                    profile  = webdriver.FirefoxProfile()
                    profile.set_proxy(proxy = proxy)
                    driver = webdriver.Remote(
                            browser_profile = profile,
                            desired_capabilities = desired_cap,
                            command_executor = command_executor)
                else:
                    driver = webdriver.Remote(
                            command_executor = command_executor,
                            desired_capabilities = desired_cap
                    )

                self.info_log('Got a session')

            except Exception as e:
                if unicode(e).find("Error forwarding the new session") != -1:
                    if retry:

                        self.info_log("Waiting 5 sec because the pool doesn't contain the needed browser.")

                        sleep(5)

                        return self.init_driver(retry = (retry - 1))
                    else:
                        raise InitDriverException("Cannot get the driver")

        #Wrap the driver in the brome proxy driver and return it
        return ProxyDriver(
            driver = driver,
            test_instance = self,
            runner = self._runner
        )

    def delete_state(self, state_pickle = None):
        """This will delete the pickle that hold the saved test state
            
        Args:
            state_pickle (path): the path of the pickle that will be delete
        """
        if not state_pickle:
            state_pickle = self.get_state_pickle_path()

        if os.path.isfile(state_pickle):
            os.remove(state_pickle)
            self.info_log("State deleted: %s"%state_pickle)

    def get_state_pickle_path(self):
        """Return the state's pickle path

        The project:url config need to be set in order to extract the server name. Each server will have it own state
        The test name is used to name the state.

        E.g.: the test named "test 1" that run with the project:url ("http://www.example.com") with have a state named:
        tests/states/test_1_example.com.pkl

        Returns: str (path)
        """

        #Extract the server name
        server = urlparse(self.pdriver.get_config_value("project:url")).netloc

        states_dir = os.path.join(
            self.get_config_value("project:absolute_path"),
            self.get_config_value("project:script_folder_name"),
            "states"
        )
        create_dir_if_doesnt_exist(states_dir)

        #TODO should be configurable
        state_pickle_path = os.path.join(
            states_dir,
            string_to_filename('%s_%s.pkl'%(self._name.replace(' ', '_'), server))
        )

        return  state_pickle_path

    def save_state(self):
        """Save the state in a pickle

        First remove all the instance variables that are private (_*).
        Then remove the pdriver instance variable.
        Finally cleanup the state using the Stateful.cleanup_state function.
        The cleanup is recursive.

        The test state will be saved to the path returned by the get_state_pickle_path.
        
        Returns: None
        """
        self.info_log("Saving state...")

        state = {}

        #Remove all the instance variables that are private
        state = {key:value for (key, value) in self.__dict__.iteritems() if key[0] != '_'}

        #Remove the pdriver
        del state['pdriver']

        #Cleanup the state
        effective_state = Stateful.cleanup_state(state)

        #Save the state
        state_pickle_path = self.get_state_pickle_path()
        with open(state_pickle_path, 'wb') as fd:
            pickle.dump(effective_state, fd)

        self.info_log("State saved: %s"%state_pickle_path)

    def load_state(self):
        """Load the state pickle into the instance object if a state is found

        The state will be loaded in the basetest instance if a state exist.
        The pdriver will be set on each instance that inherited from Stateful
        These instance can be in a dict or a list. It is not recursive tho.

        If the instance has an after_load method, this method will be called
        E.g.:
        class Base(object):
            def after_load(self):
                self.app = self.pdriver.app

        Returns: bool True if a state was found; False if no state was found;
        """
        self.info_log("Loading state...")

        #Set the pdriver for an object that inherited from Stateful
        def set_pdriver(value):
            #TODO support class that inherited Stateful
            if Stateful in value.__class__.__bases__:
                value.pdriver = self.pdriver
                if hasattr(value, 'after_load'):
                    value.after_load()
            elif type(value) is list:
                for v in value:
                    set_pdriver(v)
            elif type(value) is dict:
                for k, v in value.iteritems():
                    set_pdriver(v)

        #Load the state pickle
        state_pickle_path = self.get_state_pickle_path()
        if os.path.isfile(state_pickle_path):
            with open(state_pickle_path, 'rb') as fd:
                state = pickle.load(fd)
        else:
            self.info_log("No state found.")

            return False

        #Set the pdriver
        for key, value in state.iteritems():
            set_pdriver(value)

        #Update the instance variable dictionary
        self.__dict__.update(state)

        self.info_log("State loaded.")

        return True

    def configure_logger(self):
        """Configure the logger for the test

        Configure the logger from the logger_test config:
            logger_test:level
            logger_test:streamlogger
            logger_test:filelogger
            logger_test:format

        See documentation: http://brome.readthedocs.org/en/release/configuration.html#logger-test

        If the config project:test_batch_result_path is set to False then the filelogger will not be configured.

        Returns: None
        """

        #Logger name
        logger_name = self._name

        #Log directory
        if self._runner_dir:
            self.test_log_dir = os.path.join(
                self._runner_dir,
                "logs"
            )

        #Logger
        self._logger = logging.getLogger(logger_name)

        #Create the log directory
        if self._runner_dir:
            create_dir_if_doesnt_exist(self.test_log_dir)

        #Format
        format_ = self.get_config_value("logger_test:format")

        #Stream logger 
        if self.get_config_value('logger_test:streamlogger'):
            sh = logging.StreamHandler()
            stream_formatter = logging.Formatter(format_)
            sh.setFormatter(stream_formatter)
            self._logger.addHandler(sh)

        #File logger
        if self._runner_dir:
            if self.get_config_value('logger_test:filelogger'):
                test_name = string_to_filename(self._name)
                fh = logging.FileHandler(os.path.join(
                    self.test_log_dir,
                    "%s_%s.log"%(test_name, self._browser_config.get_id())
                ))
                file_formatter = logging.Formatter(format_)
                fh.setFormatter(file_formatter)
                self._logger.addHandler(fh)

        #Set level
        self._logger.setLevel(getattr(logging, self.get_config_value('logger_test:level')))

    def get_logger_dict(self):
        return {'batchid': self._test_batch_id, 'testname': u"%s"%self._name}

    def debug_log(self, msg):
        self._logger.debug(u"[debug]%s"%msg, extra=self.get_logger_dict())

    def info_log(self, msg):
        self._logger.info(u"%s"%msg, extra=self.get_logger_dict())

    def warning_log(self, msg):
        self._logger.warning(u"[warning]%s"%msg, extra=self.get_logger_dict())

    def error_log(self, msg):
        self._logger.error(u"[error]%s"%msg, extra=self.get_logger_dict())

    def critical_log(self, msg):
        self._logger.critical(u"[critical]%s"%msg, extra=self.get_logger_dict())

    def execute(self):
        try:
            self.before_run()

            if self._test_config.get("delete_state"):
                self.delete_state(self.get_state_pickle_path())

            if not self.load_state():
                if hasattr(self, 'create_state'):
                    self.create_state()
                    self.save_state()

            self.run(**self._test_config)

            self.after_run()

        except Exception, e:
            self.error_log('Test failed')

            tb = traceback.format_exc()

            self.error_log('Crash: %s'%tb)

            self.fail(tb)
        finally:
            self.end()

    def end(self):
        self.info_log("Test ended")

        self.quit_driver()

        self.stop_video_recording()

        if self.get_config_value("runner:play_sound_on_test_finished"):
            say(self.get_config_value("runner:sound_on_test_finished"))

        ending_timestamp = datetime.now()
        session = Session()
        sa_test_instance =  session.query(TestInstance).filter(TestInstance.id == self._test_instance_id).one()
        sa_test_instance.ending_timestamp = ending_timestamp
        session.commit()
        session.close()

        self.info_log("Ending timestamp: %s"%ending_timestamp)

    def quit_driver(self):
        self.info_log("Quitting the browser...")
        try:
            self.pdriver.quit()
        except Exception as e:
            self.error_log('Exception driver.quit(): %s'%str(e))

        """
        if hasattr(self, 'browsermobserver'):
            with open("temp.har", "w") as fd:
                fd.write(json.dumps(self.proxy.har))

            self.browsermobserver.stop()
        """

    def before_run(self):
        pass

    def after_run(self):
        pass

    def fail(self, tb):
        if self.get_config_value("runner:play_sound_on_test_crash"):
            say(self.get_config_value("runner:sound_on_test_crash"))

        if self.get_config_value("runner:embed_on_test_crash"):
            self.pdriver.embed()

        self._crash_error = '[!]%s %s crashed: %s'%(self._name, self._browser_config.get_id(), str(tb))

        self.create_crash_report(tb)

    def create_crash_report(self, tb):
        self.info_log('Creating a crash report')

        crash_name = "%s - %s"%(self.pdriver.get_id(join_char = '_'), self._name)

        extra_data = ''
        extra_data_dict = {}

        #JAVASCRIPT ERROR
        extra_data_dict['javascript_error'] = self.pdriver.get_javascript_error()
        
        #NETWORK CAPTURE
        if self._browser_config.get('enable_proxy'):
            extra_data_dict['network_capture_path'] = os.path.join(self._network_capture_relative_dir, string_to_filename('%s.data'%self._name))

        if extra_data_dict:
            extra_data = json.dumps(extra_data_dict)

        crash_screenshot_path = os.path.join(
            self._crash_report_dir,
            string_to_filename('%s.png'%crash_name)
        )

        crash_screenshot_relative_path = os.path.join(
            self._crash_report_relative_dir,
            string_to_filename('%s.png'%crash_name)
        )

        if self._runner_dir:
            #CRASH LOG
            with open(os.path.join(self._crash_report_dir, string_to_filename('%s.log'%crash_name)), 'w') as f:
                f.write(str(tb))

            #CRASH SCREENSHOT
            self.pdriver.take_screenshot(screenshot_path = crash_screenshot_path)

        #CRASH OBJECT
        session = Session()

        sa_test_crash = TestCrash(
            timestamp = datetime.now(),
            trace = str(tb),
            browser_id = self.pdriver.get_id(),
            screenshot_path = crash_screenshot_relative_path,
            videocapture_path = self._video_capture_file_relative_path,
            extra_data = extra_data,
            title = crash_name,
            test_instance_id = self._test_instance_id,
            test_batch_id = self._test_batch_id
        )

        session.add(sa_test_crash)
        session.commit()
        session.close()

    def get_config_value(self, config_name):
        if not hasattr(self, 'browser_brome_config'):
            self._browser_brome_config = parse_brome_config_from_browser_config(self._browser_config.config)

        config_list = [
            self._browser_brome_config,
            self._runner.config,
            self._runner.brome_config
        ]
        value = get_config_value(config_list, config_name)

        if hasattr(self, '_logger'):
            self.debug_log("config_value (%s): %s"%(config_name, value))

        return value

    def configure_test_result_dir(self):

        if not self._runner_dir:
            return

        #CRASH DIRECTORY
        self._crash_report_dir = os.path.join(
            self._runner_dir,
            'crashes'
        )
        create_dir_if_doesnt_exist(self._crash_report_dir)

        self._crash_report_relative_dir = os.path.join(
            self._runner.relative_runner_dir,
            'crashes'
        )

        #ASSERTION SCREENSHOT DIRECTORY
        self._assertion_screenshot_relative_dir = os.path.join(
            self._runner.relative_runner_dir,
            'assertion_screenshots',
            self.pdriver.get_id(join_char = '_')
        )

        self._assertion_screenshot_dir = os.path.join(
            self._runner_dir,
            'assertion_screenshots',
            self.pdriver.get_id(join_char = '_')
        )
        create_dir_if_doesnt_exist(self._assertion_screenshot_dir)

        #SCREENSHOT DIRECTORY
        self._screenshot_dir = os.path.join(
            self._runner_dir,
            'screenshots',
            self.pdriver.get_id(join_char = '_')
        )
        create_dir_if_doesnt_exist(self._screenshot_dir)

        #VIDEO RECORDING DIRECTORY
        if self._browser_config.get('record_session'):
            self._video_recording_dir = os.path.join(
                self._runner_dir,
                'video_recording',
                self.pdriver.get_id(join_char = '_')
            )
            create_dir_if_doesnt_exist(self._video_recording_dir)

            self._video_recording_relative_dir = os.path.join(
                self._runner.relative_runner_dir,
                'video_recording',
                self.pdriver.get_id(join_char = '_')
            )

        #NETWORK CAPTURE DIRECTORY
        if self._browser_config.get('enable_proxy'):
            self._network_capture_dir = os.path.join(
                self._runner_dir,
                'network_capture'
            )
            create_dir_if_doesnt_exist(self._network_capture_dir)
            self._network_capture_relative_dir = os.path.join(
                self._runner.relative_runner_dir,
                'network_capture'
            )

    def get_test_result_summary(self):
        results = []

        session = Session()

        base_query = session.query(TestResult).filter(TestResult.test_instance_id == self._test_instance_id).filter(TestResult.browser_id == self.pdriver.get_id())
        total_test = base_query.count()
        total_test_successful = base_query.filter(TestResult.result == True).count()
        total_test_failed = base_query.filter(TestResult.result == False).count()
        failed_tests = base_query.filter(TestResult.result == False).all()

        results.append('Total_test: %s; Total_test_successful: %s; Total_test_failed: %s'%(total_test, total_test_successful, total_test_failed))

        for failed_test in failed_tests:
            query = session.query(Test).filter(Test.id == failed_test.test_id)
            if query.count():
                test = query.one()
                test_id = test.test_id
            else:
                test_id = 'n/a'

            results.append('[%s] %s'%(test_id, failed_test.title))

        session.close()

        return results



# -*- coding: utf-8 -*-

from time import sleep
import subprocess

from flask import Blueprint, render_template, flash, redirect, url_for, request, send_from_directory, g
import flask_sijax
from flask.ext.login import login_required
from IPython import embed

from brome.webserver.testbatch.forms import LaunchForm, ReportForm
from brome.core.model.utils import *
from brome.webserver import data_controller

blueprint = Blueprint("testbatch", __name__, url_prefix='/tb',
                      static_folder="../static")

def delete_test_batch(obj_response, testbatch_id):
    data_controller.delete_test_batch(blueprint.app, testbatch_id)
    flash("The test batch (%s) have been deleted!"%testbatch_id, 'success')
    obj_response.script("window.location = '%s';"%url_for("testbatch.list"))

def stop_test_batch(obj_response, testbatch_id):
    data_controller.stop_test_batch(blueprint.app, testbatch_id)
    flash("The test batch (%s) will be stop as soon as possible..."%testbatch_id, 'success')
    obj_response.script("window.location = '%s';"%url_for("testbatch.list"))

@blueprint.route("/vnc/<string:host>")
@login_required
def vnc(host):
    #TODO make this configurable
    return render_template("testbatch/vnc.html", title = 'VNC', host = host, port = 5900, password = '1asdf!!')

@blueprint.route("/file/<path:filename>")
@login_required
def test_batch_report_file(filename):
    return send_from_directory(blueprint.app.brome.get_config_value('project:test_batch_result_path'), filename)

@flask_sijax.route(blueprint, "/network_capture/<int:testbatch_id>")
@login_required
def network_capture(testbatch_id):
    def analyse(obj_response, network_capture_name, network_capture_path, analyse_function):
        analysis = data_controller.analyse_network_capture(blueprint.app, network_capture_path, analyse_function)

        obj_response.script("$('#%s > div[name=\"result\"]').append('<div><p>Result:</p>%s</div>')"%(network_capture_name, analysis))

    if g.sijax.is_sijax_request:
        g.sijax.register_callback('analyse', analyse)
        return g.sijax.process_request()

    data = {}
    data['network_capture_list'] = data_controller.get_network_capture(blueprint.app, testbatch_id)

    return render_template(
        "testbatch/network_capture.html",
        testbatch_id = testbatch_id,
        data = data
    )

@flask_sijax.route(blueprint, "/report/<object_type>/<object_id>/", methods=['GET', 'POST'])
@login_required
def report(object_type, object_id):
    form = ReportForm(blueprint.app, object_id, object_type)

    def analyse(obj_response, network_capture_name, network_capture_path, analyse_function):
        analysis = data_controller.analyse_network_capture(blueprint.app, network_capture_path, analyse_function)

        obj_response.script("$('#%s > p > textarea[name=\"network_analysis\"]').val('%s')"%(network_capture_name, analysis))

    if g.sijax.is_sijax_request:
        g.sijax.register_callback('analyse', analyse)
        return g.sijax.process_request()

    if request and request.method in ("PUT", "POST"):
        success, msg = form.report(request.form)
        if success:
            flash("The issue has been report!", 'success')
            sleep(2)
            return redirect(url_for('testbatch.list'))
        else:
            flash(msg, 'warning')

    return render_template("testbatch/report.html", form = form)

@blueprint.route("/launch/", methods=['GET', 'POST'])
@login_required
def launch():
    form = LaunchForm(blueprint.app)

    if request and request.method in ("PUT", "POST"):
        success, msg = form.start_test_batch(request.form)
        if success:
            flash("The test batch has been started!", 'success')
            sleep(2)
            return redirect(url_for('testbatch.list'))
        else:
            flash(msg, 'warning')

    return render_template("testbatch/launch.html", form = form)

@flask_sijax.route(blueprint, "/detail/<int:testbatch_id>")
@login_required
def detail(testbatch_id):
    def update_info(obj_response, testbatch_id, interval_id, runner_log_length, current_progress):
        test_batch = data_controller.get_test_batch_detail(blueprint.app, testbatch_id)
        test_batch_log = data_controller.get_test_batch_log(blueprint.app, testbatch_id)

        if runner_log_length < len(test_batch_log):
            obj_response.script("""
                    var logs = "%s".split("|");

                    logs.forEach(function(log) {
                        $('#runnerlog').append('<h6>' + log + '</h6>');
                    });
            """%("|".join(test_batch_log[runner_log_length:])))
        else:
            if test_batch.ending_timestamp:
                obj_response.script("clearInterval(%s);"%interval_id)

        if test_batch.ending_timestamp:
            total_execution_time = data_controller.get_total_execution_time(blueprint.app, testbatch_id)

            obj_response.script("$('#testprogressdiv').remove();")
            obj_response.script("$('#testexecutiontimespan > strong').html('%s');"%total_execution_time)
            obj_response.script("$('#testexecutiontimediv').show();")
        else:
            progress = int(float(test_batch.total_finished_tests) / float(test_batch.total_tests) * 100)
            if str(progress) != current_progress.replace('%', ''):
                obj_response.script("$('#testprogress').puiprogressbar('option', 'value', %s);"%progress)

            obj_response.script("$('#total_crashes').html(%s)"%test_batch.total_crashes)
            obj_response.script("$('#total_executing_tests').html(%s)"%test_batch.total_executing_tests)
            obj_response.script("$('#total_finished_tests').html(%s)"%test_batch.total_finished_tests)
            obj_response.script("$('#total_screenshots').html(%s)"%test_batch.total_screenshots)
            obj_response.script("$('#total_test_results').html(%s)"%test_batch.total_test_results)
            obj_response.script("$('#total_failed_tests').html(%s)"%test_batch.total_failed_tests)

    if g.sijax.is_sijax_request:
        g.sijax.register_callback('delete_test_batch', delete_test_batch)
        g.sijax.register_callback('stop_test_batch', stop_test_batch)
        g.sijax.register_callback('update_info', update_info)
        return g.sijax.process_request()

    data = {}
    data['test_batch'] = data_controller.get_test_batch_detail(blueprint.app, testbatch_id)
    data['logs'] = data_controller.get_test_batch_test_instance_log(blueprint.app, testbatch_id, 0)
    data['runner_log'] = data_controller.get_test_batch_log(blueprint.app, testbatch_id)

    show_video_capture = blueprint.app.brome.get_config_value("webserver:SHOW_VIDEO_CAPTURE")
    show_test_instances = blueprint.app.brome.get_config_value("webserver:SHOW_TEST_INSTANCES")
    show_network_capture = blueprint.app.brome.get_config_value("webserver:SHOW_NETWORK_CAPTURE")

    return render_template(
        "testbatch/detail.html",
        testbatch_id = testbatch_id,
        data = data,
        show_test_instances = show_test_instances,
        show_video_capture = show_video_capture,
        show_network_capture = show_network_capture
    )

@blueprint.route("/screenshot/<int:testbatch_id>")
@login_required
def screenshot(testbatch_id):
    data = {}
    data['screenshot_list'] = data_controller.get_test_batch_screenshot(blueprint.app, testbatch_id)

    return render_template("testbatch/screenshot.html", testbatch_id = testbatch_id, data = data)

@blueprint.route("/test_instances/<int:testbatch_id>")
@login_required
def test_instances(testbatch_id):
    data = {}
    data['test_instance_list'] = data_controller.get_test_instance_list(testbatch_id)

    return render_template("testbatch/test_instances.html", testbatch_id = testbatch_id, data = data)

@blueprint.route("/video_player")
@login_required
def video_player():
    data = {}
    data['title'] = request.args.get('video_title', '')
    data['path'] = request.args.get('video_path', '')
    data['time'] = request.args.get('time', 0)

    return render_template("testbatch/video_player.html", data = data)

@blueprint.route("/video_recording_list/<int:testbatch_id>")
@login_required
def video_recording_list(testbatch_id):
    data = {}
    data['video_recording_list'] = data_controller.get_test_batch_video_recording(blueprint.app, testbatch_id)

    return render_template("testbatch/video_recording.html", testbatch_id = testbatch_id, data = data)

@blueprint.route("/testresult/<int:testbatch_id>")
@login_required
def testresult(testbatch_id):
    data = {}
    data['result_list'] = data_controller.get_test_batch_test_result(blueprint.app, testbatch_id)
    test_batch_is_running = data_controller.get_test_batch(testbatch_id).ending_timestamp == None

    return render_template("testbatch/testresult.html", testbatch_id = testbatch_id, test_batch_is_running = test_batch_is_running, data = data)

@blueprint.route("/crash/<int:testbatch_id>")
@login_required
def crash(testbatch_id):
    data = {}
    data['crash_list'] = data_controller.get_test_batch_crashes(blueprint.app, testbatch_id)

    return render_template("testbatch/crash.html", testbatch_id = testbatch_id, data = data)

@flask_sijax.route(blueprint, "/list")
@login_required
def list():
    if g.sijax.is_sijax_request:
        g.sijax.register_callback('delete_test_batch', delete_test_batch)
        g.sijax.register_callback('stop_test_batch', stop_test_batch)
        return g.sijax.process_request()

    data = {}
    data['testbatch_list'] = data_controller.get_test_batch_list()

    return render_template("testbatch/list.html", data = data)



#! -*- coding: utf-8 -*-

import traceback

from brome.core.model.utils import *
from brome.core.model.meta.base import Session
from brome.core.runner.localhost_instance import LocalhostInstance
from brome.core.runner.base_runner import BaseRunner
from brome.core.runner.browser_config import BrowserConfig
from brome.core.model.test_batch import TestBatch

class LocalRunner(BaseRunner):
    """The local runner only run browser on localhost
    """

    def __init__(self, *args):
        super(LocalRunner, self).__init__(*args)

    def execute(self):
        """Execute the test batch
            
        """

        self.browser_config = BrowserConfig(
            runner = self,
            browser_id = self.get_config_value("runner:localhost_runner"),
            browsers_config = self.brome.browsers_config
        )

        try:
            self.run()
        except KeyboardInterrupt:
            self.info_log("Test batch interrupted")

        except:
            tb = traceback.format_exc()
            self.error_log("Exception in run of the grid runner: %s"%str(tb))
            raise
        
        finally:
            self.terminate()

    def run(self):
        """Run the test batch
        """

        self.info_log("The test batch is ready.")

        self.executed_tests = []

        for test in self.tests:
            
            localhost_instance = LocalhostInstance(
                runner = self,
                browser_config = self.browser_config,
                test_name = test.Test.name
            )

            localhost_instance.startup()

            test_ = test.Test(
                runner = self,
                browser_config = self.browser_config,
                name = test.Test.name,
                test_batch_id = self.test_batch_id,
                localhost_instance = localhost_instance,
                index = 1
            )

            test_.execute()
            self.executed_tests.append(test_)

            localhost_instance.tear_down()

    def terminate(self):
        """Terminate the test batch
        """

        self.info_log('The test batch is finished.')

        session = Session()
        sa_test_batch = Session.query(TestBatch).filter(TestBatch.id == self.test_batch_id).one()
        sa_test_batch.ending_timestamp = datetime.now()
        session.commit()
        session.close()

        self.print_test_summary(self.executed_tests)



# -*- coding: utf-8 -*-
from flask_wtf import Form
from wtforms import TextField, PasswordField
from wtforms.validators import DataRequired

from brome.core.model.user import User
from brome.webserver.extensions import db

class LoginForm(Form):
    username = TextField('Username', validators=[DataRequired()])
    password = PasswordField('Password', validators=[DataRequired()])

    def __init__(self, *args, **kwargs):
        super(LoginForm, self).__init__(*args, **kwargs)
        self.user = None

    def validate(self):
        initial_validation = super(LoginForm, self).validate()
        if not initial_validation:
            return False

        self.user = db.session.query(User).filter_by(username = self.username.data).first()
        if not self.user:
            self.username.errors.append('Unknown username')
            return False

        if not self.user.check_password(self.password.data):
            self.password.errors.append('Invalid password')
            return False

        return True



#! -*- coding: utf-8 -*-

from brome.core.model.meta import SurrogatePK, Base, Column, Boolean, Text, relationship

class Test(SurrogatePK, Base):

    test_id = Column(Text())
    name = Column(Text())
    test_results = relationship("TestResult", backref="test")



#! -*- coding: utf-8 -*-

from brome.core.model.meta import SurrogatePK, Base, Column, DateTime, relationship, Integer, Boolean

class TestBatch(SurrogatePK, Base):

    pid = Column(Integer())
    killed = Column(Boolean(), default = False)
    total_tests = Column(Integer())
    starting_timestamp = Column(DateTime())
    ending_timestamp = Column(DateTime())
    test_instances = relationship("TestInstance", backref="testbatch")
    test_results = relationship("TestResult", backref="testbatch")



from sqlalchemy import String, Numeric, Integer, Column, ForeignKey, DateTime, Boolean, Text
from sqlalchemy.orm import relationship
from sqlalchemy.orm.exc import MultipleResultsFound

from brome.core.model.meta.base import setup_database, Session, Base, update_test, create_database, delete_database
from brome.core.model.meta.schema import SurrogatePK



#! -*- coding: utf-8 -*-

__version__ = "0.1.5"

import webbrowser
import shutil
import pickle
import hashlib
import os
import argparse
import shutil
import re
from glob import glob
from threading import Timer

import yaml
from tornado.wsgi import WSGIContainer
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop

from core.model.utils import *
from core.model.meta import Session
from core.model.grep import grep_files
from core.runner.local_runner import LocalRunner
from core.runner.grid_runner import GridRunner
from core.model.meta import create_database, delete_database, setup_database, update_test
from core.model.configurator import load_brome_config, get_config_value, default_config, generate_brome_config
from webserver.app import create_app

class Brome(object):
    def __init__(self, **kwargs):
        self.config_path = kwargs.get('config_path')
        self.selector_dict = kwargs.get('selector_dict', {})
        self.test_dict = kwargs.get('test_dict', {})
        self.browsers_config_path = kwargs.get('browsers_config_path')
        self.config = load_brome_config(self.config_path)
        self.config['project']['absolute_path'] = kwargs.get('absolute_path')

        with open(self.browsers_config_path, 'r') as fd:
            self.browsers_config = yaml.load(fd)

        if not self.browsers_config:
            print 'You must provide a browsers config dict to the brome instance'
            exit(1)

    def print_usage(self):
        print '$ ./bro admin | run | webserver | list | find'
        exit(1)

    def execute(self, args):
        if len(args) < 2:
            self.print_usage()

        if args[1] == 'run':
            self.run(args[2:])
        elif args[1] == 'admin':
            self.admin(args[2:])
        elif args[1] == 'find':
            self.find(args[2:])
        elif args[1] == 'list':
            self.list_(args[2:])
        elif args[1] == 'webserver':
            self.webserver(args[2:])
        else:
            self.print_usage()

    def run(self, args):
        parser = argparse.ArgumentParser(description='Brome runner')

        #TEST SEARCH QUERY
        parser.add_argument(
                            '--search',
                            '-s',
                            dest = 'test_search_query', 
                            default = '*',
                            help = 'The search query used to activate tests'
        )
        #TEST NAME
        parser.add_argument(
                            '--name',
                            '-n',
                            dest = 'test_name', 
                            help = 'The test name to be executed'
        )

        #TEST FILE
        parser.add_argument(
            '--test-file',
            dest = 'test_file',
            default = False,
            help = 'The file containing the name of all the tests that you want to run'
        )

        #LOCALHOST RUNNER
        parser.add_argument(
                            '--localhost',
                            '-l',
                            dest = 'localhost_runner', 
                            help = 'Launch a browser from localhost'
        )

        #REMOTE RUNNER
        parser.add_argument(
                            '-r',
                            '--remote',
                            dest = 'remote_runner', 
                            default = False,
                            help = 'Launch a browser on a remote'
        )

        def test_config_string(value):
            try:
                re.match("([^=]+=[^,]+,?)+", value).group(0)
                return value
            except AttributeError:
                print "--test-config '%s' does not match required format (key=value,key1:value1)"%(value)
                exit(1)

        #TEST CONFIG KWARGS
        parser.add_argument(
                            '--test-config',
                            type = test_config_string,
                            dest = 'test_config', 
                            help = 'The config that will be pass to the test. ex: "key=value,key1=value1"'
        )

        def brome_config_string(value):
            try:
                re.match("([^:]+:[^=]+=[^,]+,?)+", value).group(0)
                return value
            except AttributeError:
                print "--brome-config '%s' does not match required format (section:key=value,section1:key1=value1)"%(value)
                exit(1)

        #BROME CONFIG KWARGS
        parser.add_argument(
                            '--brome-config',
                            type = brome_config_string,
                            dest = 'brome_config', 
                            help = 'The config that will be pass to the brome runner. ex: "section:key=value,section1:key1=value1"'
        )

        self.parsed_args = parser.parse_args(args)

        if self.test_dict:
            self.auto_update_test()

        if self.parsed_args.localhost_runner:
            browsers_id = [self.parsed_args.localhost_runner]
        elif self.parsed_args.remote_runner:
            browsers_id = self.parsed_args.remote_runner.split(',')
        else:
            print 'Please select -r or -l'
            exit(1)

        for browser_id in browsers_id:
            if browser_id not in self.browsers_config.keys():
                print 'This browser id is not available in the provided browsers config'
                print 'Supported browser(s) are: %s'%self.browsers_config.keys()
                exit(1)

        #LOCALHOST RUNNER 
        if self.parsed_args.localhost_runner:
            LocalRunner(self).execute()

        #REMOTE RUNNER
        elif self.parsed_args.remote_runner:
            GridRunner(self).execute()

        #ERROR
        else:
            print 'Select either -l "{browser_id}" or -r "{browser_id}"'

    def admin(self, args):
        parser = argparse.ArgumentParser(description='Brome admin')

        parser.add_argument(
                            '--generate-config',
                            dest = 'generate_config', 
                            action = 'store_true',
                            help = 'Generate the default brome config'
        )

        parser.add_argument(
                            '--reset',
                            dest = 'reset', 
                            action = 'store_true',
                            help = 'Reset the database + delete the test batch results + update the test table + delete all the test state'
        )

        parser.add_argument(
                            '--create-database',
                            dest = 'create_database', 
                            action = 'store_true',
                            help = 'Create the project database'
        )

        parser.add_argument(
                            '--delete-test-states',
                            dest = 'delete_test_states', 
                            action = 'store_true',
                            help = 'Delete all the test states'
        )

        parser.add_argument(
                            '--delete-test-results',
                            dest = 'delete_test_result', 
                            action = 'store_true',
                            help = 'Delete all the test batch results'
        )

        parser.add_argument(
                            '--reset-database',
                            dest = 'reset_database', 
                            action = 'store_true',
                            help = 'Reset the project database'
        )

        parser.add_argument(
                            '--delete-database',
                            dest = 'delete_database', 
                            action = 'store_true',
                            help = 'Delete the project database'
        )

        parser.add_argument(
                            '--update-test',
                            dest = 'update_test', 
                            action = 'store_true',
                            help = 'Update the test in the database'
        )

        parsed_args = parser.parse_args(args)

        def reset_database():
            delete_database(self.get_config_value('database:sqlalchemy.url'))
            create_database(self.get_config_value('database:sqlalchemy.url'))

        def delete_test_states():
            states_dir = os.path.join(
                self.get_config_value("project:absolute_path"),
                "tests",
                "states"
            )
            try:
                shutil.rmtree(states_dir)
            except OSError:
                pass

            try:
                os.makedirs(states_dir)
            except OSError:
                pass

            print 'States deleted'

        def delete_test_batch_result():
            if os.path.exists(self.get_config_value('project:test_batch_result_path')):
                shutil.rmtree(self.get_config_value('project:test_batch_result_path'))
                print 'Test batch result (%s) deleted!'%self.get_config_value('project:test_batch_result_path')
            else:
                print 'Nothing to delete'

        if parsed_args.create_database:
            create_database(self.get_config_value('database:sqlalchemy.url'))
        elif parsed_args.reset:
            reset_database()
            delete_test_batch_result()
            if self.test_dict:
                self.update_test()
            delete_test_states()
        elif parsed_args.generate_config:
            generate_brome_config(self.config_path)
        elif parsed_args.delete_test_result:
            delete_test_batch_result()
        elif parsed_args.delete_test_states:
            delete_test_states()
        elif parsed_args.reset_database:
            delete_database(self.get_config_value('database:sqlalchemy.url'))
            create_database(self.get_config_value('database:sqlalchemy.url'))
        elif parsed_args.delete_database:
            delete_database(self.get_config_value('database:sqlalchemy.url'))
        elif parsed_args.update_test:
            self.update_test()

    def open_browser(self):
        if self.get_config_value("webserver:open_browser"):
            webbrowser.open("http://%s:%s/tb/list"%(self.get_config_value("webserver:HOST"), self.get_config_value("webserver:PORT")))

    def webserver(self, args):
        parser = argparse.ArgumentParser(description='Brome webserver')

        #TORNADO
        parser.add_argument(
                            '--tornado',
                            '-t',
                            dest = 'tornado', 
                            action = 'store_true',
                            help = 'The brome webserver will run on Tornado. (http://www.tornadoweb.org/en/stable/index.html#)'
        )

        parsed_args = parser.parse_args(args)

        app = create_app(self)

        Timer(2, self.open_browser).start()

        if parsed_args.tornado:
            http_server = HTTPServer(WSGIContainer(app))
            http_server.listen(self.get_config_value("webserver:PORT"))
            IOLoop.instance().start()
        else:
            app.run(host = self.get_config_value("webserver:HOST"), port = self.get_config_value("webserver:PORT"))

    def list_(self, args):
        query = os.path.join(
                        self.get_config_value("project:absolute_path"),
                        "tests",
                        "test_*.py"
                    )

        tests = glob(query)
        print "[index]\t|test name|"
        for index, test in enumerate(tests):
            test_name = test.split(os.sep)[-1][5:-3]
            print "[%s]\t%s"%(index, test_name)

    def find(self, args):
        parser = argparse.ArgumentParser(description='Brome find')

        parser.add_argument(
                            '--test-id',
                            dest = 'test_id', 
                            help = 'Find a test_id in model and test directory'
        )

        parser.add_argument(
                            '--selector',
                            dest = 'selector', 
                            help = 'Find a selector in model and test directory'
        )

        parsed_args = parser.parse_args(args)

        if parsed_args.test_id:
            pattern = "\.(assert_.*|create_test_result)+\(.*[\'\"]+(%s)+[\'\"]+"%parsed_args.test_id
        elif parsed_args.selector:
            pattern = "\..*\(+[\'\"]+.*(%s)+.*[\'\"]+"%parsed_args.selector

        paths = [
            os.path.join(
                self.get_config_value("project:absolute_path"),
                "tests"
            ),
            os.path.join(
                self.get_config_value("project:absolute_path"),
                "model"
            ),
        ]

        grep_files(paths, pattern, True)

    def get_config_value(self, config_name):
        config_list = [
            self.config
        ]
        value = get_config_value(config_list,config_name)

        return value

    def update_test(self):
        setup_database(self.get_config_value('database:*'))

        session = Session()

        if self.test_dict:
            update_test(session, self.test_dict)
        else:
            raise Exception("No test dictionary provided")

        Session.remove()
        session.close()

    def auto_update_test(self):
        old_hash = None

        brome_memory_pickle_path = os.path.join(
            self.get_config_value("project:absolute_path"),
            ".brome.pkl"
        )

        new_hash = hashlib.sha1(self.test_dict.__repr__()).hexdigest()

        if os.path.isfile(brome_memory_pickle_path): 
            with open(brome_memory_pickle_path, "rb") as fd:
                data = pickle.load(fd)

            if data.has_key('test_dict_hash'):
                if data['test_dict_hash'] != new_hash:
                    self.update_test()

                    data['test_dict_hash'] = new_hash
                    with open(brome_memory_pickle_path, "wb") as fd:
                        pickle.dump(data, fd)
        else:
            data = {}
            data['test_dict_hash'] = new_hash

            with open(brome_memory_pickle_path, "wb") as fd:
                pickle.dump(data, fd)



from sqlalchemy import Column, Integer, ForeignKeyConstraint

class SurrogatePK(object):
    """A mixin that adds a surrogate integer 'primary key' column named
    ``id`` to any declarative-mapped class."""

    id = Column(Integer, primary_key=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = []
with open('requirements.txt', 'r') as fd:
    for line in fd:
        requirements.append(line)

setup(
    name='brome',
    version='0.1.5',
    description="Framework For Selenium",
    long_description=readme,
    author="Brome-HQ",
    author_email='brome.hq@gmail.com',
    url='https://github.com/brome-hq/brome',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    license="ISCL",
    zip_safe=False,
    keywords='brome',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ]
)



#! -*- coding: utf-8 -*-

from subprocess import Popen

import paramiko
import boto.ec2
from selenium import webdriver
from selenium.webdriver.chrome.options import Options

from brome.core.model.utils import *
from .base_instance import BaseInstance

class EC2Instance(BaseInstance):
    """EC2 instance

    Attributes:
        runner (object)
        browser_config (object)
        index (int)
    """

    def __init__(self, runner, browser_config, index):
        self.runner = runner
        self.browser_config = browser_config
        self.index = index
        
    def get_ip(self):
        """Return the ip address of the node
        """

        return self.private_ip

    def execute_command(self, command, read_output = True):
        """Execute a command on the node

        Args:
            command (str)
        """

        self.info_log("executing command: %s"%command)

        try:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            k = paramiko.RSAKey.from_private_key_file(self.browser_config.get('ssh_key_path'))
            ssh.connect(self.private_ip, username = self.browser_config.get('username'), pkey = k)

            stdin, stdout, stderr = ssh.exec_command(command)

            if read_output:
                output = stdout.read()
            else:
                output = None

            ssh.close()

            return output

        except Exception as e:
            msg = "Execute_command exception: %s"%str(e)
            self.error_log(msg)
            raise Exception(msg)

    def startup(self):
        """Startup the ec2 instance
        """

        if not self.browser_config.get('launch'):
            self.warning_log("Skipping launch")
            return
        
        self.info_log("Starting up")

        instance = None
        try:

            #KEY NAME
            key_name = self.browser_config.get("ssh_key_path").split(os.sep)[-1][:-4]
            
            #SECURITY GROUP
            if type(self.browser_config.get("security_group_ids")) in [str, unicode]:
                security_group_ids = [self.browser_config.get("security_group_ids")]

            elif type(self.browser_config.get("security_group_ids")) == list:
                security_group_ids = self.browser_config.get("security_group_ids")

            else:
                msg = "The config security_group_ids must be a string or a list of string"
                self.critial_log(msg)
                raise Exception(msg)
                
            #LAUNCH INSTANCE
            ec2 = boto.ec2.connect_to_region(self.browser_config.get("region"))
            reservation = ec2.run_instances(
                    self.browser_config.get('amiid'),
                    key_name = key_name,
                    instance_type = self.browser_config.get("instance_type"),
                    security_group_ids = security_group_ids
            )

            wait_after_instance_launched = self.runner.get_config_value("ec2:wait_after_instance_launched")
            if wait_after_instance_launched:
                self.info_log("Waiting after instance launched: %s seconds..."%wait_after_instance_launched)
                sleep(wait_after_instance_launched)

            else:
                self.warning_log("Skipping waiting after instance launched")

            try:
                instance = reservation.instances[0]

            except Exception as e:
                self.critical_log('Instance reservation exception: %s'%str(e))
                raise

            self.instance_id = instance.id

            self.info_log('Waiting for the instance to start...')

            for i in range(60*5):
                try:
                    status = instance.update()
                    if status == 'running':
                        break

                    else:
                        sleep(1)
                except Exception as e:
                    self.error_log('Exception while wait pending: %s'%str(e))
                    sleep(1)

            #Wait until instance is running
            status = instance.update()
            if status == 'running':
                instance.add_tag("Name","%s-selenium-node-%s-%s"%(self.browser_config.get('platform'), self.browser_config.get('browserName'), self.index))

                self.info_log(
                    "New instance (%s) public ip (%s) private ip (%s)"%(
                        instance.id,
                        instance.ip_address,
                        instance.private_ip_address
                    )
                )
            else:
                msg = "Instance status is %s and should be (running)"%status
                self.error_log(msg)
                raise Exception(msg)

            if self.runner.get_config_value("ec2:wait_until_system_and_instance_check_performed"):
                check_successful = False

                for i in range(5*60):

                    try:

                        if not i%60:
                            if not type(status) in [unicode, str]:
                                self.info_log('System_status: %s, instance_status: %s'%(status.system_status, status.instance_status))

                        status = ec2.get_all_instance_status(instance_ids=[instance.id])[0]
                        if status.system_status.status == u'ok' and status.instance_status.status == u'ok':

                            self.info_log('system_status: %s, instance_status: %s'%(status.system_status, status.instance_status))
                            check_successful = True
                            break

                    except Exception as e:
                        self.error_log('Waiting instance ready exception: %s'%str(e))
                    sleep(1)

                if not check_successful:
                    msg = "System and instance check were not successful"
                    self.warning_log(msg)
                    raise Exception(msg)
            else:
                self.warning_log("Skipping wait until system and instance check performed")

            self.info_log('Starting the selenium node server')

            self.private_ip = instance.private_ip_address
            self.public_dns = instance.public_dns_name
            self.private_dns = instance.private_dns_name
            self.public_ip = instance.ip_address

            #LINUX
            if self.browser_config.get('platform').lower() == "linux":
                command = self.browser_config.get("selenium_command").format(**self.browser_config.config)
                self.execute_command(command)

            elif self.browser_config.get('platform').upper() == "windows":

                #TODO this code is out of date
                config = self.browser_config.config.copy()
                config['instance_ip_address'] = instance.ip_address
                command = self.browser_config("selenium_command").format(**config)
                process = Popen(command.split(" "), stdout=devnull, stderr=devnull)
                self.runner.xvfb_pids.append(process.pid)

            else:

                msg = "The provided platform name is not supported: select either (WINDOWS) or (LINUX)"
                self.critical_log(msg)
                raise Exception(msg)

            #PROXY
            if self.browser_config.get('enable_proxy'):
                port = self.browser_config.get('proxy_port', 8080)
                self.start_proxy(port = port)

            return True

        except Exception as e:
            self.error_log('Startup exception: %s'%str(e))
            raise

    def tear_down(self):
        """Tear down the instance
        """
        
        if not self.browser_config.get('terminate'):
            self.warning_log("Skipping terminate")
            return

        self.info_log("Tearing down...")
        
        ec2 = boto.ec2.connect_to_region(self.browser_config.get("region"))
        ec2.terminate_instances(instance_ids=[self.instance_id])

        #PROXY
        if self.browser_config.config.get('enable_proxy'):
            self.stop_proxy()

    def start_proxy(self, port = None):
        """Start the mitmproxy
        """
        
        self.runner.info_log("Starting proxy...")

        self.proxy_port = port
        
        self.network_data_path = os.path.join(
            self.runner.runner_dir,
            'network_data'
        )
        create_dir_if_doesnt_exist(self.network_data_path)

        self.local_proxy_output_path = os.path.join(
            self.network_data_path,
            string_to_filename('%s.data'%self.index)
        )

        self.remote_proxy_output_path = string_to_filename('%s.data'%self.index)

        path_to_mitmproxy = self.browser_config.get("mitmproxy:path", 'mitmdump')

        filter_ = self.browser_config.get("mitmproxy:filter")
        command = [
            'DISPLAY=:0',
            'nohup',
            path_to_mitmproxy,
            "-p",
            "%i"%self.proxy_port,
            "-w",
            self.remote_proxy_output_path
        ]

        if filter_:
            command.append(filter_)

        command.append('&')

        self.execute_command(' '.join(command), read_output = False)

    def stop_proxy(self):
        """Stop the mitmproxy
        """

        self.runner.info_log("Stopping proxy...")

        #scp the network data
        scp_command = [
            'scp',
            '%s@%s:%s'%(self.browser_config.get('username'), self.get_ip(), self.remote_proxy_output_path),
            self.local_proxy_output_path
        ]
        p = subprocess.Popen(scp) 
        p.wait()

        self.new_proxy_output_path = os.path.join(
            self.network_data_path,
            string_to_filename('%s.data'%self.index)
        )

        os.rename(self.local_proxy_output_path, self.new_proxy_output_path)
        os.remove(self.local_proxy_output_path)

        #kill the proxy
        self.execute_command("fuser -k %i/tcp"%self.proxy_port)

    def get_id(self):
        return '%s - %s'%(self.browser_config.browser_id, self.index)

    def debug_log(self, msg):
        self.runner.debug_log("[%s]%s"%(self.get_id(), msg))

    def info_log(self, msg):
        self.runner.info_log("[%s]%s"%(self.get_id(), msg))

    def warning_log(self, msg):
        self.runner.warning_log("[%s]%s"%(self.get_id(), msg))

    def error_log(self, msg):
        self.runner.error_log("[%s]%s"%(self.get_id(), msg))

    def critial_log(self, msg):
        self.runner.critial_log("[%s]%s"%(self.get_id(), msg))




class InitDriverException(Exception):
    """Raise when the driver failed initializing
    Args:
        msg (str): Human readable string describing the exception.
    """

    def __init__(self, msg):
        self.msg = msg

class InvalidBrowserName(Exception):
    """Raise when the given browserName config is invalid
    Args:
        msg (str): Human readable string describing the exception.
    """

    def __init__(self, msg):
        self.msg = msg



# -*- coding: utf-8 -*-

from flask import Blueprint, render_template, flash, redirect, url_for, request
from flask.ext.login import login_required
from IPython import embed

from brome.webserver.admin.forms import ConfigForm

blueprint = Blueprint("admin", __name__, url_prefix='/admin',
                      static_folder="../static")

@blueprint.route("/")
@login_required
def members():
    return render_template("admin/members.html")

@blueprint.route("/configuration/", methods=['GET', 'POST'])
@login_required
def configuration():
    form = ConfigForm(blueprint.app)

    if request and request.method in ("PUT", "POST"):
        flash("The brome configuration has been updated successfully!", 'success')
        form.save(request.form)
        return redirect(url_for('public.home'))

    return render_template("admin/configuration.html", form = form)



# -*- coding: utf-8 -*-

import subprocess
import os
import sys
import json
from datetime import datetime

import yaml
from IPython import embed

from brome.webserver import data_controller

class LaunchForm(object):
    def __init__(self, app):
        self.app = app

        self.init_data()

    def init_data(self):
        self.data = {}

        self.data['browser_list'] = data_controller.get_browser_list(self.app)

        self.data['test_list'] = data_controller.get_test_list(self.app)

    def start_test_batch(self, data):
        runner_path = os.path.join(
            self.app.brome.get_config_value('project:absolute_path'),
            'bro'
        )

        requested_browsers = [b[len("browser_"):] for b in dict(data).keys() if b.startswith("browser_")]

        if not len(requested_browsers):
            return False, 'You need to select at least one browser'

        requested_tests = [t[len("test_"):] for t in dict(data).keys() if t.startswith("test_")]

        if not len(requested_tests):
            return False, 'You need to select at least one test'

        test_file_path = os.path.join(
            self.app.temp_path,
            'test_file.yaml'
        )
        with open(test_file_path, 'w') as f:
            f.write(yaml.dump(requested_tests, default_flow_style=False))

        command = [
            sys.executable,
            runner_path,
            "run",
            "-r",
            ",".join(requested_browsers),
            "--test-file",
            test_file_path
        ]
        self.app.logger.info("Starting test bach with the following command: %s"%command)

        subprocess.Popen(
                command,
                stdout=open(os.devnull, 'w'),
                stderr=open('runner.log', 'a'),
        )

        return True, ''

class ReportForm(object):
    def __init__(self, app, object_id, object_type):
        self.app = app
        self.object_id = object_id
        self.object_type = object_type

        self.fetch_object()
        self.init_data()

    def fetch_object(self):
        if self.object_type == 'test_result':
            self.data_object = data_controller.get_test_result(self.app, self.object_id)
        elif self.object_type == 'test_crash':
            self.data_object = data_controller.get_test_crash(self.app, self.object_id)

    def init_data(self):
        self.data = {}

        #TITLE
        self.data['title'] = self.data_object.title

        #JAVASCRIPT ERROR
        try:
            self.data['javascript_error'] = json.loads(self.data_object.extra_data)['javascript_error']
        except (ValueError, KeyError):
            self.data['javascript_error'] = ''
            
        #SCREENSHOT
        self.data['screenshot_path'] = self.data_object.screenshot_path

        #NETWORK CAPTURE
        try:
            self.data['network_capture_path'] = json.loads(self.data_object.extra_data)['network_capture_path']

            if self.app.brome.get_config_value("webserver:analyse_network_capture_report_func"):
                self.data['network_capture_analyse'] = True
        except (ValueError, KeyError):
            self.data['network_capture_path'] = ''

        #VIDEO
        self.data['video_path'] = self.data_object.videocapture_path
        self.data['video_title'] = self.data_object.title.replace('_', ' ')

        test_instance = data_controller.get_test_instance(self.data_object.test_instance_id)
        self.data['video_time_position'] =  (self.data_object.timestamp - test_instance.starting_timestamp).total_seconds()

        #CUSTOM FIELDS
        self.data['custom_fields'] = self.app.brome.get_config_value("webserver:report")['custom_fields']

    def report(self, data):
        self.app.logger.info("Reported!")

        on_submit = self.app.brome.get_config_value("webserver:report")['on_submit']
        module_name = on_submit.split(':')[0]
        function_name = on_submit.split(':')[1]

        module = __import__(module_name, fromlist = [''])

        success, msg = getattr(module, function_name)(dict(data))

        return success, msg
